from typing import TYPE_CHECKING, AsyncGenerator, Optional, Sequence
import httpx
from agentia.history import History
from agentia.llm import LLMOptions
from agentia.llm.completion import ChatCompletion
from agentia.llm.stream import ChatCompletionEvents, ChatCompletionStream
from agentia.models import *
from agentia.tools.tools import ToolSet

if TYPE_CHECKING:
    from agentia.agent import Agent
    from agentia.live import Live
    from agentia.live import LiveOptions


async def __process_tool_calls(
    agent: "Agent", tool_calls: list[ToolCall], tools: ToolSet, parallel: bool
) -> tuple[ToolMessage, list[ToolCallResponse]]:
    assert tools is not None, "No tools provided"
    tool_responses = await tools.run(agent, tool_calls, parallel)
    tool_msg_parts: list[MessagePartToolResult] = []
    for tr in tool_responses:
        mptr = MessagePartToolResult(
            tool_call_id=tr.tool_call_id,
            tool_name=tr.tool_name,
            input=tr.input,
            output=tr.output,
            output_files=tr.output_files,
        )
        tool_msg_parts.append(mptr)
    tool_msg = ToolMessage(content=tool_msg_parts)
    return tool_msg, tool_responses


async def __add_prompt(
    agent: "Agent", history: History, prompt: str | Message | Sequence[Message]
) -> bool:
    """Add the prompt to history. Returns False if the prompt was a slash command
    that produced no output, in which case the LLM should be skipped."""

    prompt = await agent.commands.process_messages(prompt)

    if not prompt:
        return False

    if isinstance(prompt, str):
        history.add(UserMessage(content=[MessagePartText(text=prompt)]))
    elif not isinstance(prompt, (list, Sequence)):
        history.add(prompt)
    else:
        history.add(*prompt)
    return True


def run_react_loop(
    agent: "Agent",
    prompt: str | Message | Sequence[Message],
    options: LLMOptions,
    live_options: Optional["LiveOptions"],
) -> ChatCompletion:
    tools = agent.tools

    async def gen() -> AsyncGenerator[AssistantMessage | ToolMessage, None]:
        if not await __add_prompt(agent, agent.history, prompt):
            return
        messages = agent.history.get()
        instructions = agent.history.get_instructions()

        async with httpx.AsyncClient() as client:
            while True:
                result = await agent.provider.generate(
                    instructions=instructions,
                    messages=messages,
                    tools=tools,
                    options=options,
                    client=client,
                )
                c.usage += result.usage
                c.last_usage = result.usage
                c._add_new_message(result.message)
                messages.append(result.message)
                yield result.message
                # Add new messages
                tool_calls = result.message.tool_calls
                if result.finish_reason != "tool-calls":
                    break
                # Call tools and continue
                tool_msg, _ = await __process_tool_calls(
                    agent,
                    tool_calls,
                    tools=tools,
                    parallel=options.parallel_tool_calls or False,
                )
                c._add_new_message(tool_msg)
                messages.append(tool_msg)
                yield tool_msg
                # Process enqueued messages if any
                if msgs := agent.clear_deferred_messages():
                    for m in msgs:
                        c._add_new_message(m)
                        messages.append(m)

        agent.history.add(*c.new_messages)

    async def gen_wrapper() -> AsyncGenerator[AssistantMessage | ToolMessage, None]:
        from agentia.tools.mcp import MCPContext
        from agentia.live import LiveOptions

        await agent._auto_compact_if_needed(options)

        async with MCPContext() as _ctx:
            agent._temp_mcp_context = _ctx

            await tools.init()

            if agent.provider.supports_live:
                await agent.provider.connect_live(
                    live_options or LiveOptions(),
                    tools,
                    agent.history.get_instructions(),
                    history=agent.history,
                )

            async for msg in gen():
                yield msg
            agent._temp_mcp_context = None

            if agent.provider.supports_live:
                await agent.provider.disconnect_live()

    c = ChatCompletion(gen_wrapper(), agent)
    return c


def run_react_loop_streamed(
    agent: "Agent",
    prompt: str | Message | Sequence[Message],
    events: bool,
    options: LLMOptions,
    live_options: Optional["LiveOptions"],
) -> ChatCompletionStream | ChatCompletionEvents:
    tools = agent.tools

    async def gen() -> AsyncGenerator[StreamPart, None]:
        if not await __add_prompt(agent, agent.history, prompt):
            return
        messages = agent.history.get()
        last_finish_reason: FinishReason = "unknown"
        started = False
        instructions = agent.history.get_instructions()

        async with httpx.AsyncClient() as client:
            while True:
                tool_calls: list[ToolCall] = []
                parts: list[AssistantMessagePart] = []
                last_msg = ""
                last_reasoning = ""
                async for part in agent.provider.stream(
                    instructions=instructions,
                    messages=messages,
                    tools=tools,
                    options=options,
                    client=client,
                ):
                    if not started:
                        started = True
                        yield StreamPartStart()

                    assert part.type not in (
                        "stream-start",
                        "stream-end",
                    ), f"Provider must not yield {part.type}"

                    match part.type:
                        case "text-start":
                            last_msg = ""
                        case "text-delta":
                            last_msg += part.delta
                        case "text-end":
                            parts.append(MessagePartText(text=last_msg))
                        case "reasoning-start":
                            last_reasoning = ""
                        case "reasoning-delta":
                            last_reasoning += part.delta
                        case "reasoning-end":
                            parts.append(MessagePartReasoning(text=last_reasoning))
                        case "tool-call":
                            parts.append(
                                MessagePartToolCall(
                                    tool_call_id=part.tool_call_id,
                                    tool_name=part.tool_name,
                                    input=part.input,
                                    provider_executed=part.provider_executed,
                                )
                            )
                            if not part.provider_executed:
                                tool_calls.append(part)

                    if part.type == "turn-end":
                        s.usage += part.usage
                        s.last_usage = part.usage
                        last_finish_reason = part.finish_reason
                    else:
                        yield part
                msg = AssistantMessage(content=parts)
                yield StreamPartTurnEnd(
                    role="assistant",
                    message=msg,
                    usage=s.usage,
                    finish_reason=last_finish_reason,
                )
                messages.append(msg)
                s._add_new_message(msg)
                if not tool_calls:
                    break
                # Call tools and continue
                yield StreamPartTurnStart(role="tool")
                tool_msg, tool_responses = await __process_tool_calls(
                    agent,
                    tool_calls,
                    tools,
                    options.parallel_tool_calls or False,
                )
                for tr in tool_responses:
                    yield tr
                yield StreamPartTurnEnd(role="tool", message=tool_msg)
                messages.append(tool_msg)
                s._add_new_message(tool_msg)
                # Process enqueued messages if any
                if msgs := agent.clear_deferred_messages():
                    for m in msgs:
                        s._add_new_message(m)
                        messages.append(m)
        s.finish_reason = last_finish_reason
        yield StreamPartEnd(usage=s.usage, finish_reason=last_finish_reason)

        agent.history.add(*s.new_messages)

    async def gen_wrapper() -> AsyncGenerator[StreamPart, None]:
        from agentia.tools.mcp import MCPContext
        from agentia.live import LiveOptions

        await agent._auto_compact_if_needed(options)

        async with MCPContext() as _ctx:
            agent._temp_mcp_context = _ctx

            await tools.init()

            if agent.provider.supports_live:
                await agent.provider.connect_live(
                    live_options or LiveOptions(),
                    tools,
                    agent.history.get_instructions(),
                    history=agent.history,
                )

            async for part in gen():
                yield part
            agent._temp_mcp_context = None

            if agent.provider.supports_live:
                await agent.provider.disconnect_live()

    if events:
        s = ChatCompletionEvents(gen_wrapper(), agent)
    else:
        s = ChatCompletionStream(gen_wrapper(), agent)
    return s


async def run_react_loop_live(live: "Live") -> AsyncGenerator[StreamPart, None]:
    started = False
    parts: list[AssistantMessagePart] = []
    tool_calls: list[ToolCall] = []
    last_msg = ""
    last_reasoning = ""
    agent = live.agent

    async for part in agent.provider.receive():
        if not started:
            started = True
            yield StreamPartStart()
        assert part.type not in (
            "stream-start",
            "stream-end",
        ), f"Provider must not yield {part.type}"

        match part.type:
            case "turn-start":
                yield part
            case "text-start":
                last_msg = ""
            case "text-delta":
                last_msg += part.delta
            case "text-end":
                parts.append(MessagePartText(text=last_msg))
            case "reasoning-start":
                last_reasoning = ""
            case "reasoning-delta":
                last_reasoning += part.delta
            case "reasoning-end":
                parts.append(MessagePartReasoning(text=last_reasoning))
            case "tool-call":
                parts.append(
                    MessagePartToolCall(
                        tool_call_id=part.tool_call_id,
                        tool_name=part.tool_name,
                        input=part.input,
                        provider_executed=part.provider_executed,
                    )
                )
                if not part.provider_executed:
                    tool_calls.append(part)
        yield part

        if part.type == "turn-end":
            if part.usage.total_tokens is not None:
                agent.history.current_tokens = part.usage.total_tokens
            agent.history.usage += part.usage

        if part.type == "turn-end" or part.type == "tool-call":
            # append the assistant message to history
            msg = AssistantMessage(content=parts)
            agent.history.add(msg)
            # run tool calls if any
            if tool_calls:
                yield StreamPartTurnStart(role="tool")
                tool_msg, tool_responses = await __process_tool_calls(
                    agent, tool_calls, agent.tools, False
                )
                for tr in tool_responses:
                    yield tr
                agent.history.add(tool_msg)
                yield StreamPartTurnEnd(role="tool", message=tool_msg)
                await agent.provider.send_tool_responses(tool_responses)
            # Process enqueued messages if any
            if msgs := agent.clear_deferred_messages():
                for m in msgs:
                    agent.history.add(m)
                    await live.send_message(m)
            # reset for next turn
            parts = []
            tool_calls = []
            last_msg = ""
            last_reasoning = ""



from typing import override
from ._openai_api import OpenAIAPIProvider
import os
import httpx

# Model list: https://vercel.com/ai-gateway/models

_context_length_cache: dict[str, int] = {}


class Vercel(OpenAIAPIProvider):
    def __init__(self, model: str, api_key: str | None = None):
        api_key = (
            api_key
            or os.environ.get("VERCEL_AI_GATEWAY_API_KEY")
            or os.environ.get("VERCEL_API_KEY")
        )
        if not api_key:
            raise ValueError(
                "VERCEL_AI_GATEWAY_API_KEY or VERCEL_API_KEY environment variable not set"
            )
        base_url = "https://ai-gateway.vercel.sh/v1"
        super().__init__(name="vercel", model=model, api_key=api_key, base_url=base_url)
        self.extra_body["modalities"] = ["text", "image"]

    @override
    async def _fetch_context_length(self) -> int:
        if self.model in _context_length_cache:
            return _context_length_cache[self.model]
        async with httpx.AsyncClient() as client:
            resp = await client.get(
                f"{self.base_url}/models",
                headers={"Authorization": f"Bearer {self.api_key}"},
            )
            resp.raise_for_status()
            for model in resp.json().get("data", []):
                mid = model.get("id")
                ctx = model.get("context_window")
                if mid and ctx is not None:
                    _context_length_cache[mid] = int(ctx)
            if self.model in _context_length_cache:
                return _context_length_cache[self.model]
            raise ValueError(
                f"Context length not available for model '{self.model}' from Vercel API"
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Sequence

from agentia.models import CompactedMessage, Message
from agentia.models.chat import MessagePartText

if TYPE_CHECKING:
    from agentia.agent import Agent

EFFORT_INSTRUCTIONS = {
    "low": "Keep as much important information as possible. Only drop clearly unimportant or redundant content.",
    "medium": "Keep important information and context, but drop routine details, verbose outputs, and redundant exchanges.",
    "high": "Be very aggressive in compacting. Only keep the most critical information and drop everything else.",
}


def _format_messages(messages: Sequence[Message]) -> str:
    lines: list[str] = []
    for msg in messages:
        role = msg.role
        if isinstance(msg.content, str):
            lines.append(f"[{role}]: {msg.content}")
        elif isinstance(msg.content, Sequence):
            parts = []
            for part in msg.content:
                if isinstance(part, MessagePartText):
                    parts.append(part.text)
            if parts:
                lines.append(f"[{role}]: {''.join(parts)}")
    return "\n\n".join(lines)


async def compact_history(
    agent: Agent,
    effort: Literal["low", "medium", "high"] = "low",
    model: str | None = None,
    keep_last_n: int = 1,
) -> None:
    from agentia.agent import Agent

    messages = agent.history.get()
    length = len(messages)
    if length <= keep_last_n:
        return

    compact_instructions = (
        "Below is a conversation history. Your task is to compact it into a single concise summary "
        "that preserves all important information (key decisions, facts, code changes, conclusions, "
        "user preferences, and action items) while dropping unimportant details.\n\n"
        f"{EFFORT_INSTRUCTIONS[effort]}\n\n"
        "Output ONLY the compacted summary, nothing else."
    )

    conversation_text = _format_messages(messages[:-keep_last_n])

    compact_agent = Agent(model=model or agent.model, instructions=compact_instructions)
    result = await compact_agent.run(conversation_text)

    messages_to_keep = messages[-keep_last_n:]
    agent.history.clear()
    agent.history.add(
        CompactedMessage(content=result.text, original_messages=list(messages)),
        *messages_to_keep,
    )



from . import Plugin, tool
import asyncio
import os
import signal
import tempfile
import weakref
from typing import Annotated, override
from pathlib import Path
from ..tools import ToolResult
from ..models.base import File
import base64
from asyncio.subprocess import DEVNULL, PIPE, STDOUT


_DEFAULT_MAX_LINES = 1000
_DEFAULT_MAX_BYTES = 256 * 1024
_DEFAULT_HEAD_RATIO = 0.3


def _terminate_background_pids(pids: set[int]) -> None:
    for pid in list(pids):
        try:
            os.kill(pid, signal.SIGTERM)
        except Exception:
            pass


_FILE_MIME_TYPES = {
    ".png": "image/png",
    ".jpg": "image/jpeg",
    ".jpeg": "image/jpeg",
    ".gif": "image/gif",
    ".webp": "image/webp",
    ".pdf": "application/pdf",
}


def _truncate(
    content: str,
    *,
    max_head_lines: int,
    max_tail_lines: int,
    save_full: bool = False,
) -> tuple[str, str | None]:  # returns (truncated_content, full_content_path)
    """
    Keep up to `max_head_lines` from the start and `max_tail_lines` from the end;
    lines in between are replaced with a `[... N lines omitted ...]` marker.
    Set `max_tail_lines=0` for head-only truncation, or `max_head_lines=0` for tail-only.
    The byte budget (`_DEFAULT_MAX_BYTES`) is split between head and tail in proportion to their line caps.
    """
    max_lines = max_head_lines + max_tail_lines
    if not content or max_lines == 0:
        return "", None
    total_bytes = len(content.encode("utf-8"))
    lines = content.split("\n")
    total_lines = len(lines)
    if total_bytes <= _DEFAULT_MAX_BYTES and total_lines <= max_lines:
        return content, None
    max_head_bytes = _DEFAULT_MAX_BYTES * max_head_lines // max_lines
    max_tail_bytes = _DEFAULT_MAX_BYTES - max_head_bytes

    def _take(reverse: bool, lim_lines: int, lim_bytes: int) -> list[str]:
        kept: list[str] = []
        used = 0
        for line in reversed(lines) if reverse else lines:
            if len(kept) >= lim_lines:
                break
            sz = len(line.encode("utf-8")) + (1 if kept else 0)
            if kept and used + sz > lim_bytes:
                break
            kept.append(line)
            used += sz
        return list(reversed(kept)) if reverse else kept

    head = _take(False, max_head_lines, max_head_bytes)
    tail = _take(True, max_tail_lines, max_tail_bytes)

    if len(head) + len(tail) >= total_lines:
        body = content
    elif not tail:
        body = "\n".join(head)
    elif not head:
        body = "\n".join(tail)
    else:
        omitted = total_lines - len(head) - len(tail)
        body = "\n".join(head + [f"[... {omitted} lines omitted ...]"] + tail)

    path: str | None = None
    if save_full:
        fd, path = tempfile.mkstemp(prefix="agentia-bash-", suffix=".log", text=True)
        with os.fdopen(fd, "w") as f:
            f.write(content)
    return body, path


class System(Plugin):
    def __init__(self, bash=True) -> None:
        super().__init__()
        self.bash_enabled = bash
        self._background_pids: set[int] = set()
        weakref.finalize(self, _terminate_background_pids, self._background_pids)

    @override
    def get_instructions(self) -> str | None:
        return f"CURRENT WORKING DIRECTORY: `{str(Path.cwd())}`"

    @tool(name="Bash")
    async def run_bash_command(
        self,
        command: Annotated[str, "The bash command to run."],
        cwd: Annotated[
            str | None, "The optional working directory for the command."
        ] = None,
        timeout: Annotated[int | None, "Optional timeout in seconds."] = 5 * 60,
        background: Annotated[
            bool,
            "If true, start the command detached and return immediately with its PID.",
        ] = False,
    ) -> str:
        """
        Run a bash command and return its combined stdout and stderr.

        **Truncation:**
            * Large output is truncated with the full output spooled to a temp file.

        **Non-zero exit code:**
            * When the command exits with a non-zero code, it is appended to the output.

        **Background / timeout:**
            * If `background=True`, the process is started detached and the call returns immediately with its PID and the path of the log file its stdout/stderr is streaming to.
            * While running background tasks, any new output is polled and enqueued regularly.
            * If `timeout` fires for a foreground command, the process is left running in the background. `timeout` is set to 5 minutes by default, and has no effect if `background=True`.
                * set timeout = 0 to disable the timeout and run indefinitely in the foreground. e.g. when running `sleep 1h`.
            * Use background tasks for long-running or daemon processes (e.g. dev servers). Only use this when necessary.
        """
        if not self.bash_enabled:
            raise RuntimeError("Bash tool is not enabled.")

        if timeout is not None and timeout <= 0:
            timeout = None

        await self.agent.user_consent_guard("Run command: " + command)

        # Stream both stdout and stderr to a single temp log file so we can
        # both return the combined output and (in background mode) tail it.
        fd, log_path = tempfile.mkstemp(prefix="agentia-bash-", suffix=".log")
        os.close(fd)
        with open(log_path, "wb") as log_write:
            proc = await asyncio.create_subprocess_shell(
                command,
                stdout=log_write,
                stderr=STDOUT,
                cwd=cwd,
                # Detach into a new session so timeouts/background mode leave
                # the process alive without us having to babysit its tty.
                start_new_session=True,
            )

        if background:
            # Hand monitoring off to a fire-and-forget task and return now.
            # The task tails `log_path` and feeds output back via agent.enqueue
            # so the model sees progress without us blocking the tool call.
            self._background_pids.add(proc.pid)
            return (
                f"[Started background process pid={proc.pid}. Output streaming to: {log_path}. "
                f"New output is enqueued regularly. Use `TaskStop` tool to terminate.]"
            )

        try:
            returncode = await asyncio.wait_for(proc.wait(), timeout=timeout)
        except TimeoutError:
            # Don't kill the process — leave it running so the user/agent can
            # decide whether to wait longer or stop it via TaskStop. Hand off
            # to the same background monitor so the agent still sees progress
            # and the eventual exit code instead of going dark.
            self._background_pids.add(proc.pid)
            return (
                f"[Command timed out after {timeout}s. Process is still running in background pid={proc.pid}. "
                f"Output streaming to: {log_path}. New output is enqueued regularly. Use `TaskStop` tool to terminate]"
            )

        output = Path(log_path).read_text(encoding="utf-8", errors="replace")
        try:
            os.unlink(log_path)
        except OSError:
            pass

        # Head-heavy split keeps the start of the output (often the most
        # diagnostic) while still preserving the tail where errors usually land.
        head_lines = int(_DEFAULT_MAX_LINES * _DEFAULT_HEAD_RATIO)
        body, full_path = _truncate(
            output,
            max_head_lines=head_lines,
            max_tail_lines=_DEFAULT_MAX_LINES - head_lines,
            save_full=True,
        )
        if not body:
            body = "(no output)"
        if full_path:
            body += f"\n\n[Output truncated. Full output saved to: {full_path}]"
        if returncode != 0:
            body += f"\n\n[Command exited with code {returncode}]"
        return body

    @tool(name="TaskStop")
    async def task_stop(
        self,
        pid: Annotated[int, "Process ID of the task to stop."],
    ) -> str:
        """
        Stop a running process by PID using the `kill` command. Sends SIGTERM first;
        if the process is still alive after a brief grace period, escalates to SIGKILL.
        Use this to terminate background processes started by `Bash` (either via `background=True` or after a timeout).
        """
        await self.agent.user_consent_guard(f"Kill process: {pid}")

        async def _kill(sig_flag: str) -> tuple[int, str]:
            proc = await asyncio.create_subprocess_exec(
                "kill", sig_flag, str(pid), stdout=PIPE, stderr=PIPE
            )
            _, stderr = await proc.communicate()
            assert proc.returncode is not None
            return proc.returncode, stderr.decode(errors="replace").strip()

        async def _alive() -> bool:
            proc = await asyncio.create_subprocess_exec(
                "kill",
                "-0",
                str(pid),
                stdout=DEVNULL,
                stderr=DEVNULL,
            )
            await proc.wait()
            return proc.returncode == 0

        rc, err = await _kill("-TERM")
        if rc != 0:
            msg = err or f"kill exited with code {rc}"
            return f"[Failed to kill pid={pid}: {msg}]"

        for _ in range(20):
            await asyncio.sleep(1)
            if not await _alive():
                return f"[Sent SIGTERM to pid={pid}]"

        rc, err = await _kill("-KILL")
        if rc != 0:
            msg = err or f"kill -9 exited with code {rc}"
            return f"[SIGTERM did not stop pid={pid}; SIGKILL also failed: {msg}]"
        self._background_pids.discard(pid)
        return f"[SIGTERM did not stop pid={pid}; sent SIGKILL]"

    @tool(name="Read")
    async def read_file(
        self,
        path: Annotated[str, "Path to the file"],
        offset: Annotated[int | None, "Start line number (1-indexed)"] = None,
        limit: Annotated[int | None, "Max number of lines to read"] = None,
    ) -> str | ToolResult:
        """
        Read the contents of a file so that you can inspect their content.
            * Text files: return the UTF-8 text content (maybe truncated).
            * Images and PDFs: return as attachments with appropriate media types.
        """
        p = Path(path)
        if not p.exists():
            raise ValueError(f"File '{path}' does not exist.")
        if p.is_dir():
            raise ValueError(f"'{path}' is a directory, not a file.")

        # Non-text files
        suffix = p.suffix.lower()
        if suffix in _FILE_MIME_TYPES:
            media_type = _FILE_MIME_TYPES[suffix]
            encoded = base64.b64encode(p.read_bytes()).decode()
            url = f"data:{media_type};base64,{encoded}"
            return ToolResult(
                files=[File(media_type=media_type, data=url)],
                output=f"Loaded file '{path}' ({media_type}).",
            )

        # Text files
        text = p.read_text(encoding="utf-8")
        all_lines = text.splitlines()
        total = len(all_lines)
        start = max(1, offset) if offset else 1
        if total > 0 and start > total:
            raise ValueError(
                f"Offset {offset} is out of range for file with {total} lines."
            )
        max_lines = min(limit, _DEFAULT_MAX_LINES) if limit else _DEFAULT_MAX_LINES
        sliced = "\n".join(all_lines[start - 1 : start - 1 + max_lines])
        body, _ = _truncate(sliced, max_head_lines=max_lines, max_tail_lines=0)
        kept = body.count("\n") + 1 if body else 0
        end = start + kept - 1
        if end < total:
            body += f"\n\n[Showed lines {start}-{end} of {total}. Pass offset={end + 1} to continue.]"
        return body

    @tool(name="Write")
    async def write_file(
        self,
        path: Annotated[str, "Path to the file"],
        content: Annotated[str, "Full content to write"],
    ) -> str:
        """
        Write `content` to `path`. Missing parent directories are created automatically. Existing content is overwritten.
        """
        await self.agent.user_consent_guard(f"Write file: {path}")
        p = Path(path)
        p.parent.mkdir(parents=True, exist_ok=True)
        p.write_text(content)
        return f"Wrote {len(content)} characters to '{path}'."

    @tool(name="Edit")
    async def edit_file(
        self,
        path: Annotated[str, "Path to the file"],
        old_text: Annotated[
            str,
            "Exact text to replace. Must appear verbatim: including whitespace and newlines, and exactly once in the file.",
        ],
        new_text: Annotated[str, "Replacement text."],
    ) -> str:
        """
        Surgically replace a unique substring in a file. `old_text` must match
        exactly and appear exactly once. The edit fails — leaving the file
        untouched — if there are zero or multiple matches, or if `new_text`
        equals `old_text`. For larger rewrites, use Write instead.
        """
        await self.agent.user_consent_guard(f"Edit file: {path}")
        p = Path(path)
        if not p.exists():
            raise ValueError(f"File '{path}' does not exist.")
        if p.is_dir():
            raise ValueError(f"'{path}' is a directory, not a file.")
        content = p.read_text()
        matches = content.count(old_text)
        if matches == 0:
            raise ValueError(
                f"`old_text` was not found in '{path}'. The match must be exact, including whitespace and newlines."
            )
        if matches > 1:
            raise ValueError(
                f"`old_text` matches {matches} places in '{path}'. Include more surrounding context so the match is unique."
            )
        if old_text == new_text:
            raise ValueError(
                f"No change applied to '{path}': `old_text` and `new_text` are identical."
            )
        p.write_text(content.replace(old_text, new_text, 1))
        return f"Edited '{path}': replaced {len(old_text)} characters with {len(new_text)}."



from typing import override
from ._openai_api import OpenAIAPIProvider
import os
import httpx

# Model list: https://openrouter.ai/models

_context_length_cache: dict[str, int] = {}


class OpenRouter(OpenAIAPIProvider):
    def __init__(self, model: str, api_key: str | None = None):
        api_key = api_key or os.environ.get("OPENROUTER_API_KEY")
        if not api_key:
            raise ValueError("OPENROUTER_API_KEY environment variable not set")
        base_url = "https://openrouter.ai/api/v1"
        super().__init__(
            name="openrouter", model=model, api_key=api_key, base_url=base_url
        )
        m = os.getenv("AGENTIA_OPENROUTER_MODALITIES", None)
        if m:
            self.extra_body["modalities"] = [
                x.strip().lower() for x in m.split(",") if x.strip()
            ]
        t = os.getenv("AGENTIA_OPENROUTER_TRANSFORMS", None)
        if t and t != "0" and t.lower() != "false":
            self.extra_body["transforms"] = ["middle-out"]
        if p := os.getenv("OPENROUTER_PROVIDERS", None):
            self.extra_body["provider"] = {
                "order": [x.strip().lower() for x in p.split(",") if x.strip()],
            }

    @staticmethod
    async def fetch_context_length(model: str) -> int:
        """Fetch context length from OpenRouter's model catalog.

        Caches all models from the API response. Can be used by other providers
        that don't expose context length in their own APIs.
        """
        if model in _context_length_cache:
            return _context_length_cache[model]
        async with httpx.AsyncClient() as client:
            resp = await client.get("https://openrouter.ai/api/v1/models")
            resp.raise_for_status()
            for m in resp.json().get("data", []):
                mid = m.get("id")
                ctx = m.get("context_length")
                if mid and ctx is not None:
                    _context_length_cache[mid] = int(ctx)
            if model in _context_length_cache:
                return _context_length_cache[model]
            raise ValueError(
                f"Context length not available for model '{model}' from OpenRouter API"
            )

    @override
    async def _fetch_context_length(self) -> int:
        return await self.fetch_context_length(self.model)



import abc
import asyncio
from typing import AsyncGenerator, Literal, TYPE_CHECKING, Optional, Sequence, overload
from pydantic import BaseModel, Field
from agentia.llm.react import run_react_loop_live
from agentia.models import StreamPart
from io import BytesIO
from agentia.models.chat import Message, MessagePartText
from agentia.utils.aec import EchoCanceller
from agentia.utils.image_scaling import ScaleOption, downscale_jpeg
from agentia.models.live import (
    LiveChunk,
    LiveChunkAudio,
    LiveChunkText,
    LiveChunkImage,
    LiveChunkVideo,
    LiveChunkEnd,
)

if TYPE_CHECKING:
    from agentia.agent import Agent


class LiveOptions(BaseModel):
    """Configuration for a Gemini Live session."""

    modalities: list[Literal["text", "audio"]] = Field(
        default_factory=lambda: ["audio"]
    )
    """Response modalities. Currently only AUDIO is reliably supported by Gemini Live API."""

    voice: str | None = None
    """Voice name (e.g. "Puck", "Charon", "Kore", "Fenrir", "Aoede")."""

    language: str | None = None
    """Language code (e.g. "en")."""

    thinking_level: Literal["minimal", "low", "medium", "high"] | None = None
    """Thinking level for native audio models. Default is "minimal" for lowest latency."""

    vad: bool = True
    """Enable voice activity detection for automatic interruption handling."""


class InputStream(abc.ABC):
    def __init__(self):
        self._live: Optional["Live"] = None

    async def init(self): ...
    async def deinit(self): ...

    async def send(self, chunk: LiveChunk) -> None:
        """Send a chunk to the input stream."""
        assert self._live is not None, "InputStream not connected to Live session"
        await self._live._input_queue.put(chunk)

    @abc.abstractmethod
    async def run(self): ...


class OutputStream(abc.ABC):
    def __init__(self):
        self._live: Optional["Live"] = None

    async def init(self): ...
    async def deinit(self): ...

    @overload
    async def receive(self, kind: Literal["audio"]) -> LiveChunkAudio | None: ...

    @overload
    async def receive(
        self, kind: Literal["text", "transcription"]
    ) -> LiveChunkText | None: ...

    async def receive(
        self, kind: Literal["text", "audio", "transcription"]
    ) -> LiveChunkAudio | LiveChunkText | None:
        """Receive a chunk from the output stream."""
        assert self._live is not None, "OutputStream not connected to Live session"
        if kind == "text":
            return await self._live._output_text_queue.get()
        elif kind == "audio":
            return await self._live._output_audio_queue.get()
        elif kind == "transcription":
            return await self._live._output_transcription_queue.get()

    @abc.abstractmethod
    async def run(self): ...


class VideoInput(InputStream):
    def __init__(
        self,
        kind: Literal["camera", "screen"],
        device: int | None = None,
        fps: float = 1,
        scale: ScaleOption = "raw",
    ):
        super().__init__()
        self.kind = kind
        self.fps = fps
        self.device = device
        self.scale: ScaleOption = scale
        self.cap = None

    async def init(self):
        if self.kind == "camera":
            import cv2

            self.cap = cv2.VideoCapture(self.device or 0)
            self.grab_camera_frame()
        else:
            self.grab_screenshot()

    async def deinit(self):
        if self.cap is not None:
            self.cap.release()

    def grab_screenshot(self) -> bytes:
        import pyscreenshot

        img = pyscreenshot.grab()
        buf = BytesIO()
        img.convert("RGB").save(buf, format="JPEG")  # type: ignore
        return buf.getvalue()

    def grab_camera_frame(self) -> bytes:
        import cv2
        from PIL import Image

        assert self.cap is not None, "Camera capture not initialized"
        ret, frame = self.cap.read()
        if not ret:
            raise RuntimeError("Failed to grab camera frame")
        img = Image.fromarray(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
        buf = BytesIO()
        img.save(buf, format="JPEG")  # type: ignore
        return buf.getvalue()

    async def run(self):
        while True:
            await asyncio.sleep(1.0 / float(self.fps))
            if self.kind == "screen":
                data = await asyncio.to_thread(self.grab_screenshot)
            else:
                data = await asyncio.to_thread(self.grab_camera_frame)
            data = await asyncio.to_thread(downscale_jpeg, data, self.scale)
            await self.send(LiveChunkVideo(data=data, mime_type="image/jpeg"))


class AudioInput(InputStream):
    def __init__(
        self,
        sample_rate: int = 16000,
        chunk_size: int = 1024,
        device: int | None = None,
        echo_canceller: EchoCanceller | None = None,
    ):
        super().__init__()

        import pyaudio

        self.pya = pyaudio.PyAudio()
        self.format = pyaudio.paInt16
        self.channels = 1
        self.sample_rate = sample_rate
        self.chunk_size = chunk_size
        self.stream = None
        self.device = device
        self.echo_canceller = echo_canceller

    async def init(self):
        self.stream = await asyncio.to_thread(
            self.pya.open,
            format=self.format,
            channels=self.channels,
            rate=self.sample_rate,
            input=True,
            frames_per_buffer=self.chunk_size,
            input_device_index=self.device,
        )

    async def deinit(self):
        if self.stream is not None:
            self.stream.close()
        self.pya.terminate()

    async def run(self):
        assert self.stream is not None, "Audio stream not initialized"
        while True:
            data = await asyncio.to_thread(
                self.stream.read, self.chunk_size, exception_on_overflow=False
            )
            if self.echo_canceller is not None:
                data = await asyncio.to_thread(self.echo_canceller.process, data)
            await self.send(LiveChunkAudio(data=data, mime_type="audio/pcm;rate=16000"))


class AudioOutput(OutputStream):
    def __init__(
        self,
        sample_rate: int = 24000,
        device: int | None = None,
        echo_canceller: EchoCanceller | None = None,
    ):
        super().__init__()

        import pyaudio

        self.pya = pyaudio.PyAudio()
        self.format = pyaudio.paInt16
        self.channels = 1
        self.sample_rate = sample_rate
        self.stream = None
        self.device = device
        self.echo_canceller = echo_canceller

    async def init(self):
        self.stream = await asyncio.to_thread(
            self.pya.open,
            format=self.format,
            channels=self.channels,
            rate=self.sample_rate,
            output=True,
            output_device_index=self.device,
        )

    async def deinit(self):
        if self.stream is not None:
            self.stream.close()
        self.pya.terminate()

    async def run(self):
        assert self.stream is not None, "Audio stream not initialized"
        while True:
            chunk = await self.receive("audio")
            if chunk is None:
                continue
            if self.echo_canceller is not None:
                self.echo_canceller.feed_reference(chunk.data)
            await asyncio.to_thread(self.stream.write, chunk.data)


class TextOutput(OutputStream):
    def __init__(self):
        super().__init__()

    async def run(self):
        while True:
            chunk = await self.receive("text")
            if chunk is None:
                print()
            else:
                print(chunk.text, end="", flush=True)


class TranscriptionOutput(OutputStream):
    def __init__(self):
        super().__init__()

    async def run(self):
        while True:
            chunk = await self.receive("transcription")
            if chunk is None:
                print()
            else:
                print(chunk.text, end="", flush=True)


class Live:
    def __init__(self, agent: "Agent", options: LiveOptions | None = None):
        self.agent = agent
        self.options = options or LiveOptions()
        self.__nesting_level = 0
        self.__multithread_mode = False
        self._input_queue = asyncio.Queue[LiveChunk]()
        self._output_audio_queue = asyncio.Queue[LiveChunkAudio | None]()
        self._output_text_queue = asyncio.Queue[LiveChunkText | None]()
        self._output_transcription_queue = asyncio.Queue[LiveChunkText | None]()

    async def __aenter__(self):
        self.__nesting_level += 1
        if self.__nesting_level > 1:
            # Already in a live session, just return self without reconnecting
            return self
        await self.agent.__aenter__()
        # Connect live session if provider supports it
        assert self.agent.provider.supports_live
        await self.agent.tools.init()
        instructions = self.agent.history.get_instructions() or None
        await self.agent.provider.connect_live(
            self.options, self.agent.tools, instructions, history=self.agent.history
        )
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.__nesting_level -= 1
        if self.__nesting_level > 0:
            # Still in a nested live session, do not disconnect yet
            return
        # Disconnect live session if provider supports it
        await self.agent.provider.disconnect_live()
        await self.agent.__aexit__(exc_type, exc_val, exc_tb)

    @overload
    async def send(self, *, chunk: LiveChunk): ...
    @overload
    async def send(self, *, text: str): ...
    @overload
    async def send(self, *, audio: bytes, mime_type: str | None = None): ...
    @overload
    async def send(self, *, image: bytes, mime_type: str | None = None): ...
    @overload
    async def send(self, *, video: bytes, mime_type: str | None = None): ...
    @overload
    async def send(self, *, end: Literal[True]): ...

    async def send(
        self,
        *,
        image: bytes | None = None,
        video: bytes | None = None,
        audio: bytes | None = None,
        text: str | None = None,
        end: Literal[True] | None = None,
        mime_type: Optional[str] = None,
        chunk: LiveChunk | None = None,
    ):
        """Send a chunk to the live session."""
        args = dict(
            image=image, video=video, audio=audio, text=text, end=end, chunk=chunk
        )

        def check_args(active: str):
            for k, v in args.items():
                if k != active:
                    assert v is None, f"`{k}` should not be provided"
            if active in ("text", "end", "chunk"):
                assert mime_type is None, "mime_type should not be specified"

        if chunk:
            check_args("chunk")
        else:
            if image:
                check_args("image")
                chunk = LiveChunkImage(data=image, mime_type=mime_type or "image/jpeg")
            elif video:
                check_args("video")
                chunk = LiveChunkVideo(data=video, mime_type=mime_type or "image/jpeg")
            elif audio:
                check_args("audio")
                chunk = LiveChunkAudio(
                    data=audio, mime_type=mime_type or "audio/pcm;rate=16000"
                )
            elif text:
                check_args("text")
                chunk = LiveChunkText(text=text)
            elif end is True:
                check_args("end")
                chunk = LiveChunkEnd()
            else:
                raise ValueError("Invalid input: no data provided")

        if self.__multithread_mode:
            await self._input_queue.put(chunk)
            return
        await self.agent.provider.send_live_chunk(chunk)

    async def send_message(self, msg: Message):
        """Send a chat message to the live session. Only supported by certain providers."""
        if not self.agent.provider.supports_live:
            raise RuntimeError("Provider does not support live sessions")
        match msg.role:
            case "user":
                parts = (
                    [MessagePartText(text=msg.content)]
                    if isinstance(msg.content, str)
                    else msg.content
                )
                for p in parts:
                    if p.type == "text":
                        await self.send(text=p.text)
                    elif p.type == "file":
                        mime = p.media_type
                        data = p.to_bytes()
                        match mime.split("/")[0]:
                            case "image":
                                await self.send(image=data, mime_type=mime)
                            case "audio":
                                await self.send(audio=data, mime_type=mime)
                            case "video":
                                await self.send(video=data, mime_type=mime)
                            case _:
                                raise NotImplementedError(
                                    f"Unsupported file type for live input: {mime}"
                                )
            case _:
                raise NotImplementedError(
                    f"Unsupported message type for live input: {type(msg)}"
                )

    async def receive(self) -> AsyncGenerator[StreamPart, None]:
        """Receive stream parts from the live session."""
        if self.__multithread_mode:
            raise RuntimeError(
                "Multithread mode does not support receive() - use start() instead"
            )
        async for event in run_react_loop_live(self):
            yield event

    async def __receive_task(self):
        async for event in run_react_loop_live(self):
            match event.type:
                case "text-delta":
                    chunk = LiveChunkText(text=event.delta)
                    await self._output_text_queue.put(chunk)
                case "text-end":
                    await self._output_text_queue.put(None)
                case "audio-delta":
                    chunk = LiveChunkAudio(
                        data=event.delta, mime_type="audio/pcm;rate=16000"
                    )
                    await self._output_audio_queue.put(chunk)
                case "audio-end":
                    await self._output_audio_queue.put(None)
                case "output-transcription-delta":
                    chunk = LiveChunkText(text=event.delta)
                    await self._output_transcription_queue.put(chunk)
                case "output-transcription-end":
                    await self._output_transcription_queue.put(None)

    async def __send_task(self):
        while True:
            chunk = await self._input_queue.get()
            await self.agent.provider.send_live_chunk(chunk)

    async def start(
        self,
        inputs: (
            Sequence[InputStream | Literal["audio", "camera", "screen"]] | None
        ) = None,
        outputs: (
            Sequence[OutputStream | Literal["text", "audio", "transcription"]] | None
        ) = None,
        camera: Optional[int] = None,
        screen: Optional[int] = None,
        scale: ScaleOption = "raw",
        mic: Optional[int] = None,
        speaker: Optional[int] = None,
        aec: bool = False,
    ):
        """Start the live session in multithread mode, allowing concurrent send/receive."""
        self.__multithread_mode = True

        # Create shared echo canceller if AEC is enabled
        echo_canceller: EchoCanceller | None = None

        def _get_echo_canceller() -> EchoCanceller:
            nonlocal echo_canceller
            if echo_canceller is None:
                echo_canceller = EchoCanceller()
            return echo_canceller

        def create_istream(kind: Literal["audio", "camera", "screen"]) -> InputStream:
            if kind == "audio":
                return AudioInput(
                    device=mic,
                    echo_canceller=_get_echo_canceller() if aec else None,
                )
            elif kind == "camera":
                return VideoInput(kind="camera", device=camera, scale=scale)
            elif kind == "screen":
                return VideoInput(kind="screen", device=screen, scale=scale)
            else:
                raise NotImplementedError(
                    f"Input stream for {kind} not implemented yet"
                )

        def create_ostream(
            kind: Literal["text", "audio", "transcription"],
        ) -> OutputStream:
            if kind == "audio":
                return AudioOutput(
                    device=speaker,
                    echo_canceller=_get_echo_canceller() if aec else None,
                )
            elif kind == "text":
                return TextOutput()
            elif kind == "transcription":
                return TranscriptionOutput()
            else:
                raise NotImplementedError(
                    f"Output stream for {kind} not implemented yet"
                )

        inputs = [
            i if isinstance(i, InputStream) else create_istream(i) for i in inputs or []
        ]
        outputs = [
            o if isinstance(o, OutputStream) else create_ostream(o)
            for o in outputs or []
        ]

        for o in outputs:
            await o.init()
        for i in inputs:
            await i.init()

        try:
            async with self:
                async with asyncio.TaskGroup() as tg:
                    tasks = []
                    for i in inputs:
                        i._live = self
                        tasks.append(i.run())
                    for o in outputs:
                        o._live = self
                        tasks.append(o.run())
                    for t in tasks:
                        tg.create_task(t)
                    tg.create_task(self.__receive_task())
                    tg.create_task(self.__send_task())
        finally:
            for o in outputs:
                await o.deinit()
            for i in inputs:
                await i.deinit()


__all__ = ["LiveOptions", "Live", "EchoCanceller"]



from typing import Callable, Sequence
from agentia.models import *


class History:
    def __init__(self) -> None:
        self.__instruction_generators: list[str | Callable[[], str | None]] = []
        self.__messages: list[Message] = []
        self.__live_cursor: int = 0
        self.current_tokens: int = 0
        self.usage: Usage = Usage()

    def add_instructions(self, instructions: str | Callable[[], str | None]) -> None:
        self.__instruction_generators.append(instructions)

    def get_instructions(self) -> str:
        return self.__get_instructions()

    def __get_instructions(self) -> str:
        instructions = []
        for gen in self.__instruction_generators:
            if callable(gen):
                result = gen()
                if result:
                    instructions.append(result)
            else:
                if gen:
                    instructions.append(gen)
        return "\n\n".join(instructions)

    def add(self, *messages: Message) -> None:
        self.__messages.extend(messages)

    def get(self) -> list[Message]:
        return list(self.__messages)

    def get_new(self) -> Sequence[Message]:
        """Return only messages added since the last cursor advance."""
        return self.__messages[self.__live_cursor :]

    def advance_cursor(self) -> None:
        """Move cursor to the end of the current message list."""
        self.__live_cursor = len(self.__messages)

    def clear(self) -> None:
        self.__messages = []
        self.__live_cursor = 0
        self.current_tokens = 0
        self.usage = Usage()

    def load(self, messages: Sequence[Message]) -> None:
        self.clear()
        for m in messages:
            self.__messages.append(m)


__all__ = ["History"]



import inspect
import shlex
from typing import TYPE_CHECKING, Awaitable, Callable, Sequence

from agentia.models.chat import Message, MessagePartText, UserMessage

if TYPE_CHECKING:
    from agentia.agent import Agent


CommandHandler = Callable[[list[str]], str | None | Awaitable[str | None]]


class Commands:
    """
    Slash command dispatcher for an `Agent`.

    Parses inputs of the form `/name arg0 arg1 ...` and dispatches to:
        1. A custom handler registered via `register(name, handler)`.
        2. Otherwise, a user-invocable skill with the same name (if the
           agent has the `Skills` plugin loaded).
        3. Otherwise, does nothing.
    """

    def __init__(self, agent: "Agent") -> None:
        self.__agent = agent
        self.__handlers: dict[str, CommandHandler] = {}

    def register(self, name: str, handler: CommandHandler) -> None:
        """Register a custom slash command handler. The leading `/` is optional."""
        self.__handlers[_strip_slash(name)] = handler

    def unregister(self, name: str) -> None:
        """Remove a previously registered custom handler. No-op if absent."""
        self.__handlers.pop(_strip_slash(name), None)

    def has(self, name: str) -> bool:
        """Return True if `name` resolves to a custom handler or a user-invocable skill."""
        n = _strip_slash(name)
        if n in self.__handlers:
            return True
        return self.__find_skill(n) is not None

    @staticmethod
    def parse(input: str) -> tuple[str, list[str]] | None:
        """
        Parse `/name arg0 arg1 ...` into `(name, [arg0, arg1, ...])`.
        Returns `None` if `input` is not a slash command or is malformed.
        """
        if not input.startswith("/"):
            return None
        try:
            tokens = shlex.split(input[1:])
        except ValueError:
            return None
        if not tokens:
            return None
        return tokens[0], tokens[1:]

    async def handle(self, input: str) -> str | None:
        """
        Dispatch a slash command. Returns the handler/skill output, or `None`
        if the input is not a slash command or no handler/skill matched.
        """
        parsed = self.parse(input)
        if parsed is None:
            return input
        name, args = parsed
        if handler := self.__handlers.get(name):
            result = handler(args)
            if inspect.isawaitable(result):
                result = await result
            if result and isinstance(result, str) and not result:
                result = None
            return result
        if skill := self.__find_skill(name):
            return skill.execute(args)
        return input  # treat as normal input if no command matched

    async def process_messages(
        self, prompt: str | Message | Sequence[Message]
    ) -> list[Message]:
        messages = []
        for p in (
            prompt
            if isinstance(prompt, Sequence) and not isinstance(prompt, str)
            else [prompt]
        ):
            if isinstance(p, str):
                if p := await self.handle(p):
                    messages.append(UserMessage(p))
            elif isinstance(p, UserMessage):
                if isinstance(p.content, str):
                    if c := await self.handle(p.content):
                        messages.append(UserMessage(c))
                else:
                    parts = []
                    for part in p.content:
                        if isinstance(part, MessagePartText):
                            if c := await self.handle(part.text):
                                parts.append(MessagePartText(c))
                        else:
                            parts.append(part)
                    if parts:
                        messages.append(UserMessage(parts))
            else:
                messages.append(p)
        return messages

    def __find_skill(self, name: str):
        from agentia.plugins.skills import Skill, Skills

        skills = self.__agent.tools.get_plugin(Skills)
        if skills is None:
            return None
        skill: Skill | None = skills.user_invocable_skills.get(name)
        return skill


def _strip_slash(name: str) -> str:
    return name[1:] if name.startswith("/") else name



import base64
import os
from typing import TYPE_CHECKING, Any, AsyncGenerator, cast, override, Optional
from uuid import uuid4

import httpx
from google import genai
from google.genai import types

if TYPE_CHECKING:
    from agentia.history import History
    from google.genai.live import AsyncSession

from agentia.live import LiveOptions
from agentia.llm import LLMOptions
from agentia.llm.providers import GenerationResult, Provider
from agentia.models.base import FunctionTool, ToolCallResponse, Usage
from agentia.models.chat import (
    AssistantMessage,
    Message,
    MessagePartFile,
    MessagePartReasoning,
    MessagePartText,
    MessagePartToolCall,
    MessagePartToolResult,
    ToolMessage,
    UserMessage,
)
from agentia.models.base import ToolCall
from agentia.models.stream import (
    StreamPart,
    StreamPartAudioDelta,
    StreamPartAudioEnd,
    StreamPartAudioStart,
    StreamPartTextDelta,
    StreamPartTextEnd,
    StreamPartTextStart,
    StreamPartInputTranscriptionDelta,
    StreamPartInputTranscriptionEnd,
    StreamPartInputTranscriptionStart,
    StreamPartOutputTranscriptionDelta,
    StreamPartOutputTranscriptionEnd,
    StreamPartOutputTranscriptionStart,
    StreamPartTurnEnd,
    StreamPartTurnStart,
)
from agentia.tools.tools import ToolSet
from agentia.models.live import (
    LiveChunk,
    LiveChunkAudio,
    LiveChunkText,
    LiveChunkImage,
    LiveChunkVideo,
    LiveChunkEnd,
)


def _get_usage(u: types.UsageMetadata | None) -> Usage:
    if not u:
        return Usage()
    return Usage(
        input_tokens=u.prompt_token_count,
        output_tokens=u.response_token_count,
        total_tokens=u.total_token_count,
        reasoning_tokens=u.thoughts_token_count,
        cached_input_tokens=u.cached_content_token_count,
    )


def _convert_tools_to_genai(tools: ToolSet) -> list[types.Tool] | None:
    schemas = tools.get_schema()
    if not schemas:
        return None
    declarations: list[types.FunctionDeclaration] = []
    for schema in schemas:
        if isinstance(schema, FunctionTool):
            declarations.append(
                types.FunctionDeclaration(
                    name=schema.name,
                    description=schema.description,
                    parameters_json_schema=schema.input_schema,
                )
            )
    if not declarations:
        return None
    return [types.Tool(function_declarations=declarations)]


def _file_part_to_blob(part: MessagePartFile) -> types.Blob:
    """Convert a MessagePartFile's data to a Gemini Blob."""
    data = part.data
    if isinstance(data, bytes):
        return types.Blob(data=data, mime_type=part.media_type)
    elif isinstance(data, str) and data.startswith(("http://", "https://")):
        raise ValueError(
            "URL-based file parts are not supported for Gemini Live; provide raw bytes or base64 data."
        )
    elif isinstance(data, str):
        if data.startswith("data:"):
            # data URL: extract base64 payload
            _, payload = data.split(",", 1)
            return types.Blob(data=base64.b64decode(payload), mime_type=part.media_type)
        else:
            # Treat as base64 string
            return types.Blob(data=base64.b64decode(data), mime_type=part.media_type)
    raise ValueError(f"Unsupported file data type: {type(data)}")


def _convert_message_to_content(msg: Message) -> types.Content | None:
    """Convert an agentia Message to a google.genai types.Content for history seeding."""
    if isinstance(msg, UserMessage):
        parts: list[types.Part] = []
        if isinstance(msg.content, str):
            parts.append(types.Part(text=msg.content))
        else:
            for p in msg.content:
                if isinstance(p, MessagePartText):
                    parts.append(types.Part(text=p.text))
                elif isinstance(p, MessagePartFile):
                    blob = _file_part_to_blob(p)
                    parts.append(types.Part(inline_data=blob))
        return types.Content(role="user", parts=parts) if parts else None

    if isinstance(msg, AssistantMessage):
        parts = []
        if isinstance(msg.content, str):
            parts.append(types.Part(text=msg.content))
        else:
            for p in msg.content:
                if isinstance(p, MessagePartText):
                    parts.append(types.Part(text=p.text))
                elif isinstance(p, MessagePartReasoning):
                    parts.append(types.Part(thought=True, text=p.text))
                elif isinstance(p, MessagePartToolCall):
                    parts.append(
                        types.Part(
                            function_call=types.FunctionCall(
                                name=p.tool_name, args=p.input
                            )
                        )
                    )
        return types.Content(role="model", parts=parts) if parts else None

    if isinstance(msg, ToolMessage):
        parts = []
        for p in msg.content:
            if isinstance(p, MessagePartToolResult):
                output = p.output
                if not isinstance(output, dict):
                    output = {"result": output}
                parts.append(
                    types.Part(
                        function_response=types.FunctionResponse(
                            name=p.tool_name, response=output
                        )
                    )
                )
        return types.Content(role="user", parts=parts) if parts else None

    return None


def _convert_messages_to_contents(
    messages: list[Message],
) -> list[types.Content]:
    """Convert a list of agentia Messages to Gemini Content objects."""
    contents: list[types.Content] = []
    for msg in messages:
        content = _convert_message_to_content(msg)
        if content:
            contents.append(content)
    return contents


async def _send_user_message_realtime(
    session: "AsyncSession", msg: UserMessage
) -> None:
    """Send a UserMessage via send_realtime_input, supporting all part types."""
    if isinstance(msg.content, str):
        await session.send_realtime_input(text=msg.content)
        return

    for p in msg.content:
        match p.type:
            case "text":
                await session.send_realtime_input(text=p.text)
            case "file":
                blob = _file_part_to_blob(p)
                if p.media_type.startswith("audio/"):
                    await session.send_realtime_input(audio=blob)
                elif p.media_type.startswith(("image/", "video/")):
                    await session.send_realtime_input(video=blob)
                else:
                    # Fallback: send as video for other binary types
                    await session.send_realtime_input(video=blob)


def _build_config(
    options: LiveOptions,
    tools: ToolSet,
    instructions: str | None,
) -> types.LiveConnectConfig:
    modalities = [
        types.Modality.AUDIO if m == "audio" else types.Modality.TEXT
        for m in options.modalities
    ]

    speech_config: types.SpeechConfig | None = None
    if options.voice or options.language:
        speech_config = types.SpeechConfig(
            voice_config=(
                types.VoiceConfig(
                    prebuilt_voice_config=types.PrebuiltVoiceConfig(
                        voice_name=options.voice
                    )
                )
                if options.voice
                else None
            ),
            language_code=options.language,
        )

    thinking_config: types.ThinkingConfig | None = None
    if options.thinking_level:
        level_map = {
            "minimal": types.ThinkingLevel.MINIMAL,
            "low": types.ThinkingLevel.LOW,
            "medium": types.ThinkingLevel.MEDIUM,
            "high": types.ThinkingLevel.HIGH,
        }
        thinking_config = types.ThinkingConfig(
            thinking_level=level_map[options.thinking_level]
        )

    input_audio_transcription = types.AudioTranscriptionConfig()
    output_audio_transcription = types.AudioTranscriptionConfig()

    context_window_compression = types.ContextWindowCompressionConfig(
        sliding_window=types.SlidingWindow()
    )

    session_resumption = None
    # session_resumption = types.SessionResumptionConfig(transparent=True)

    realtime_input_config: types.RealtimeInputConfig | None = None
    if not options.vad:
        realtime_input_config = types.RealtimeInputConfig(
            automatic_activity_detection=types.AutomaticActivityDetection(disabled=True)
        )

    genai_tools: Any = _convert_tools_to_genai(tools)

    history_config = types.HistoryConfig(initial_history_in_client_content=True)

    return types.LiveConnectConfig(
        response_modalities=modalities,
        speech_config=speech_config,
        thinking_config=thinking_config,
        system_instruction=instructions or None,
        tools=genai_tools,
        input_audio_transcription=input_audio_transcription,
        output_audio_transcription=output_audio_transcription,
        context_window_compression=context_window_compression,
        session_resumption=session_resumption,
        realtime_input_config=realtime_input_config,
        history_config=history_config,
    )


_context_length_cache: dict[str, int] = {}


class GeminiLive(Provider):
    name = "gemini-live"

    def __init__(self, model: str):
        super().__init__(name="gemini-live", model=model)
        if model != "gemini-3.1-flash-live-preview":
            raise ValueError(
                "Unsupported Gemini Live model. Currently only 'gemini-3.1-flash-live-preview' is supported."
            )
        self._session: Optional["AsyncSession"] = None
        self._session_cm: Any = None  # The async context manager from connect()
        self._client: genai.Client | None = None
        self._history: Optional["History"] = None

    def _get_client(self) -> genai.Client:
        api_key = os.environ.get("GOOGLE_API_KEY") or os.environ.get("GEMINI_API_KEY")
        if not api_key:
            raise ValueError(
                "GOOGLE_API_KEY or GEMINI_API_KEY environment variable must be set"
            )
        return genai.Client(api_key=api_key)

    @override
    async def _fetch_context_length(self) -> int:
        if self.model in _context_length_cache:
            return _context_length_cache[self.model]
        client = self._client or self._get_client()
        model_info = await client.aio.models.get(model=f"models/{self.model}")
        if model_info.input_token_limit is None:
            raise ValueError(
                f"Context length not available for model '{self.model}' from Gemini API"
            )
        _context_length_cache[self.model] = model_info.input_token_limit
        return _context_length_cache[self.model]

    @property
    @override
    def supports_live(self) -> bool:
        return True

    @override
    async def connect_live(
        self,
        options: LiveOptions,
        tools: ToolSet,
        instructions: str | None,
        history: "History",
    ) -> None:
        self._client = self._get_client()
        self._history = history
        config = _build_config(options, tools, instructions)
        self._session_cm = self._client.aio.live.connect(
            model=self.model, config=config
        )
        self._session = await self._session_cm.__aenter__()

        # Seed initial history via send_client_content
        if history:
            if initial_messages := history.get():
                contents = _convert_messages_to_contents(initial_messages)
                if contents:
                    contents = cast(list[types.Content | types.ContentDict], contents)
                    assert self._session is not None
                    await self._session.send_client_content(
                        turns=contents, turn_complete=False
                    )
            history.advance_cursor()

    @override
    async def disconnect_live(self) -> None:
        if self._session_cm:
            await self._session_cm.__aexit__(None, None, None)
            self._session_cm = None
            self._session = None
        self._client = None
        self._history = None

    def _assert_session(self) -> "AsyncSession":
        if self._session is None:
            raise RuntimeError(
                "No active live session. Use 'async with agent.live()' to start a live session."
            )
        return self._session

    @override
    async def send_live_chunk(self, chunk: LiveChunk):
        session = self._assert_session()
        match chunk:
            case LiveChunkAudio(data=data, mime_type=mime):
                await session.send_realtime_input(
                    audio=types.Blob(data=data, mime_type=mime)
                )
            case LiveChunkVideo(data=data, mime_type=mime):
                await session.send_realtime_input(
                    video=types.Blob(data=data, mime_type=mime)
                )
            case LiveChunkImage(data=data, mime_type=mime):
                await session.send_realtime_input(
                    media=types.Blob(data=data, mime_type=mime)
                )
            case LiveChunkText(text=text):
                await session.send_realtime_input(text=text)
            case LiveChunkEnd():
                await session.send_realtime_input(activity_end={})

    @override
    async def send_tool_responses(self, responses: list[ToolCallResponse]) -> None:
        session = self._assert_session()
        function_responses = [
            types.FunctionResponse(
                id=r.tool_call_id,
                name=r.tool_name,
                response=(
                    r.output if isinstance(r.output, dict) else {"result": r.output}
                ),
            )
            for r in responses
        ]
        await session.send_tool_response(function_responses=function_responses)

    @override
    async def receive(self) -> AsyncGenerator[StreamPart, None]:
        session = self._assert_session()
        turn_started = False
        text_started = False
        text_id = str(uuid4())
        audio_started = False
        audio_id = str(uuid4())
        input_tx_started = False
        input_tx_id = str(uuid4())
        output_tx_started = False
        output_tx_id = str(uuid4())
        while True:
            async for response in session.receive():
                content = response.server_content
                if content:
                    if content.model_turn and content.model_turn.parts:
                        if not turn_started:
                            yield StreamPartTurnStart(role="assistant")
                            turn_started = True
                        for part in content.model_turn.parts:
                            if part.inline_data and part.inline_data.data:
                                if not audio_started:
                                    yield StreamPartAudioStart(id=audio_id)
                                    audio_started = True
                                yield StreamPartAudioDelta(
                                    id=audio_id,
                                    delta=part.inline_data.data,
                                )
                            if part.text:
                                if not text_started:
                                    yield StreamPartTextStart(id=text_id)
                                    text_started = True
                                yield StreamPartTextDelta(id=text_id, delta=part.text)
                    if content.input_transcription and content.input_transcription.text:
                        if not input_tx_started:
                            yield StreamPartInputTranscriptionStart(id=input_tx_id)
                            input_tx_started = True
                        yield StreamPartInputTranscriptionDelta(
                            id=input_tx_id,
                            delta=content.input_transcription.text,
                        )
                    if (
                        content.output_transcription
                        and content.output_transcription.text
                    ):
                        if not output_tx_started:
                            yield StreamPartOutputTranscriptionStart(id=output_tx_id)
                            output_tx_started = True
                        yield StreamPartOutputTranscriptionDelta(
                            id=output_tx_id,
                            delta=content.output_transcription.text,
                        )
                    if content.interrupted is True:
                        if audio_started:
                            yield StreamPartAudioEnd(id=audio_id)
                            audio_started = False
                            audio_id = str(uuid4())
                        if text_started:
                            yield StreamPartTextEnd(id=text_id)
                            text_started = False
                            text_id = str(uuid4())
                        if output_tx_started:
                            yield StreamPartOutputTranscriptionEnd(id=output_tx_id)
                            output_tx_started = False
                            output_tx_id = str(uuid4())
                        yield StreamPartTurnEnd(
                            usage=_get_usage(response.usage_metadata),
                            finish_reason="interrupted",
                            role="assistant",
                            message=None,
                        )
                        turn_started = False
                    elif content.turn_complete or content.generation_complete:
                        if audio_started:
                            yield StreamPartAudioEnd(id=audio_id)
                            audio_started = False
                            audio_id = str(uuid4())
                        if text_started:
                            yield StreamPartTextEnd(id=text_id)
                            text_started = False
                            text_id = str(uuid4())
                        if input_tx_started:
                            yield StreamPartInputTranscriptionEnd(id=input_tx_id)
                            input_tx_started = False
                            input_tx_id = str(uuid4())
                        if output_tx_started:
                            yield StreamPartOutputTranscriptionEnd(id=output_tx_id)
                            output_tx_started = False
                            output_tx_id = str(uuid4())
                        yield StreamPartTurnEnd(
                            usage=_get_usage(response.usage_metadata),
                            finish_reason="stop",
                            role="assistant",
                            message=None,
                        )
                        turn_started = False

                tool_call_resp = response.tool_call
                if tool_call_resp and tool_call_resp.function_calls:
                    for fc in tool_call_resp.function_calls:
                        tc_id = fc.id or str(uuid4())
                        yield ToolCall(
                            tool_call_id=tc_id,
                            tool_name=fc.name or "",
                            input=dict(fc.args) if fc.args else {},
                        )

    # --- Request/response emulation over live session ---

    async def _send_new_messages(self, new_messages: list[Message]) -> None:
        """Send new messages that the live session hasn't seen yet."""
        if not new_messages:
            return
        session = self._assert_session()

        for msg in new_messages:
            # Send the message via realtime input (triggers model response)
            match msg.role:
                case "user":
                    await _send_user_message_realtime(session, msg)
                case "tool":
                    # ToolMessage is user-originated content, send via realtime input
                    for part in msg.content:
                        if isinstance(part, MessagePartToolResult):
                            response_data = (
                                part.output
                                if isinstance(part.output, dict)
                                else {"result": part.output}
                            )
                            await session.send_tool_response(
                                function_responses={
                                    "id": part.tool_call_id,
                                    "name": part.tool_name,
                                    "response": response_data,
                                }
                            )
                case _:
                    raise NotImplementedError(
                        f"Unsupported message type for live input: {type(msg)}"
                    )

    def _get_new_messages(self, messages: list[Message]) -> list[Message]:
        """Get messages the live session hasn't seen, using History cursor if available."""
        if self._history:
            new = self._history.get_new()
            self._history.advance_cursor()
            return list(new)
        # Fallback: treat all messages as new
        return messages

    @override
    async def generate(
        self,
        instructions: str,
        messages: list[Message],
        tools: ToolSet,
        options: LLMOptions,
        client: httpx.AsyncClient,
    ) -> GenerationResult:
        session = self._assert_session()

        await self._send_new_messages(self._get_new_messages(messages))
        await session.send_realtime_input(activity_end={})

        # Collect response
        text_parts: list[str] = []
        async for response in session.receive():
            content = response.server_content
            if content:
                if content.model_turn and content.model_turn.parts:
                    for part in content.model_turn.parts:
                        if part.text:
                            text_parts.append(part.text)
                if content.output_transcription and content.output_transcription.text:
                    text_parts.append(content.output_transcription.text)
                if (
                    content.turn_complete
                    or content.interrupted
                    or content.generation_complete
                ):
                    break

        full_text = "".join(text_parts)
        message = AssistantMessage(content=[MessagePartText(text=full_text)])
        return GenerationResult(message=message, finish_reason="stop", usage=Usage())

    @override
    async def stream(
        self,
        instructions: str,
        messages: list[Message],
        tools: ToolSet,
        options: LLMOptions,
        client: httpx.AsyncClient,
    ) -> AsyncGenerator[StreamPart, None]:
        session = self._assert_session()

        await self._send_new_messages(self._get_new_messages(messages))
        await session.send_realtime_input(activity_end={})

        text_started = False
        text_id = str(uuid4())
        started = False

        async for response in session.receive():
            if not started:
                started = True
                yield StreamPartTurnStart(role="assistant")

            content = response.server_content
            if content:
                if content.model_turn and content.model_turn.parts:
                    for part in content.model_turn.parts:
                        if part.text:
                            if not text_started:
                                yield StreamPartTextStart(id=text_id)
                                text_started = True
                            yield StreamPartTextDelta(id=text_id, delta=part.text)
                    if (
                        content.output_transcription
                        and content.output_transcription.text
                    ):
                        if not text_started:
                            yield StreamPartTextStart(id=text_id)
                            text_started = True
                        yield StreamPartTextDelta(
                            id=text_id, delta=content.output_transcription.text
                        )
                if (
                    content.turn_complete
                    or content.interrupted
                    or content.generation_complete
                ):
                    break

        if text_started:
            yield StreamPartTextEnd(id=text_id)

        yield StreamPartTurnEnd(
            usage=Usage(), finish_reason="stop", role="assistant", message=None
        )



from typing import Any, override
from ._openai_api import OpenAIAPIProvider
import os
import httpx

# Model list: https://ollama.com/library

_context_length_cache: dict[str, int] = {}


class Ollama(OpenAIAPIProvider):
    def __init__(
        self, model: str, api_key: str | None = None, base_url: str | None = None
    ):
        base_url = base_url or os.environ.get(
            "OLLAMA_BASE_URL",
            os.environ.get("OPENAI_BASE_URL", "http://localhost:11434/v1"),
        )
        api_key = api_key or os.environ.get("OLLAMA_API_KEY", "dummy")
        super().__init__(name="ollama", model=model, api_key=api_key, base_url=base_url)

    @override
    def get_reasoning_args(
        self,
        enabled: bool | None,
        effort: str | None,
        exclude: bool | None,
        max_tokens: int | None,
    ) -> dict[str, Any]:
        if not enabled:
            return {}
        return {
            "think": enabled,
        }

    @override
    async def _fetch_context_length(self) -> int:
        if self.model in _context_length_cache:
            return _context_length_cache[self.model]
        # Ollama's base_url ends with /v1; the show endpoint is at the Ollama API root
        assert self.base_url is not None
        ollama_base = self.base_url.rstrip("/").removesuffix("/v1")
        async with httpx.AsyncClient() as client:
            resp = await client.post(
                f"{ollama_base}/api/show",
                json={"name": self.model},
            )
            resp.raise_for_status()
            data = resp.json()
            model_info = data.get("model_info", {})
            for key, value in model_info.items():
                if "context_length" in key:
                    _context_length_cache[self.model] = int(value)
                    return _context_length_cache[self.model]
            raise ValueError(
                f"Context length not available for model '{self.model}' from Ollama API"
            )



from typing import Annotated, Literal, Sequence
from pydantic import BaseModel, Field

from agentia.models.chat import AssistantMessage, ToolMessage
from .base import *


class StreamPartTextStart(BaseModel):
    type: Literal["text-start"] = "text-start"
    id: str


class StreamPartTextDelta(BaseModel):
    type: Literal["text-delta"] = "text-delta"
    delta: str
    id: str


class StreamPartTextEnd(BaseModel):
    type: Literal["text-end"] = "text-end"
    id: str


class StreamPartReasoningStart(BaseModel):
    type: Literal["reasoning-start"] = "reasoning-start"
    id: str


class StreamPartReasoningDelta(BaseModel):
    type: Literal["reasoning-delta"] = "reasoning-delta"
    delta: str
    id: str


class StreamPartReasoningEnd(BaseModel):
    type: Literal["reasoning-end"] = "reasoning-end"
    id: str


class StreamPartAudioStart(BaseModel):
    type: Literal["audio-start"] = "audio-start"
    id: str


class StreamPartAudioDelta(BaseModel):
    type: Literal["audio-delta"] = "audio-delta"
    id: str
    delta: bytes
    """PCM audio data. Output is 24kHz, 16-bit, mono, little-endian."""


class StreamPartAudioEnd(BaseModel):
    type: Literal["audio-end"] = "audio-end"
    id: str


class StreamPartInputTranscriptionStart(BaseModel):
    type: Literal["input-transcription-start"] = "input-transcription-start"
    id: str


class StreamPartInputTranscriptionDelta(BaseModel):
    type: Literal["input-transcription-delta"] = "input-transcription-delta"
    id: str
    delta: str


class StreamPartInputTranscriptionEnd(BaseModel):
    type: Literal["input-transcription-end"] = "input-transcription-end"
    id: str


class StreamPartOutputTranscriptionStart(BaseModel):
    type: Literal["output-transcription-start"] = "output-transcription-start"
    id: str


class StreamPartOutputTranscriptionDelta(BaseModel):
    type: Literal["output-transcription-delta"] = "output-transcription-delta"
    id: str
    delta: str


class StreamPartOutputTranscriptionEnd(BaseModel):
    type: Literal["output-transcription-end"] = "output-transcription-end"
    id: str


class StreamPartTurnStart(BaseModel):
    type: Literal["turn-start"] = "turn-start"
    role: Literal["assistant", "tool"]


class StreamPartTurnEnd(BaseModel):
    type: Literal["turn-end"] = "turn-end"
    usage: Usage = Usage()
    finish_reason: FinishReason = "stop"
    role: Literal["assistant", "tool"]
    message: Annotated[
        AssistantMessage | ToolMessage | None, Field(discriminator="role")
    ]


class StreamPartStart(BaseModel):
    type: Literal["start"] = "start"


class StreamPartEnd(BaseModel):
    type: Literal["end"] = "end"
    usage: Usage = Usage()
    finish_reason: FinishReason = "stop"


class Annotations(BaseModel):
    type: Literal["annotations"] = "annotations"
    annotations: Sequence[Annotation]

    def __init__(
        self,
        annotations: Sequence[Annotation],
        *,
        type: Literal["annotations"] = "annotations",
    ):
        super().__init__(type=type, annotations=annotations)


type StreamPart = Annotated[
    StreamPartTextStart
    | StreamPartTextDelta
    | StreamPartTextEnd
    | StreamPartReasoningStart
    | StreamPartReasoningDelta
    | StreamPartReasoningEnd
    | ToolCall
    | ToolCallResponse
    | Annotation
    | StreamPartAudioStart
    | StreamPartAudioDelta
    | StreamPartAudioEnd
    | StreamPartInputTranscriptionStart
    | StreamPartInputTranscriptionDelta
    | StreamPartInputTranscriptionEnd
    | StreamPartOutputTranscriptionStart
    | StreamPartOutputTranscriptionDelta
    | StreamPartOutputTranscriptionEnd
    | StreamPartTurnStart
    | StreamPartTurnEnd
    | StreamPartStart
    | StreamPartEnd,
    Field(discriminator="type"),
]

__all__ = [
    "StreamPartTextStart",
    "StreamPartTextDelta",
    "StreamPartTextEnd",
    "StreamPartReasoningStart",
    "StreamPartReasoningDelta",
    "StreamPartReasoningEnd",
    "StreamPartAudioStart",
    "StreamPartAudioDelta",
    "StreamPartAudioEnd",
    "StreamPartInputTranscriptionStart",
    "StreamPartInputTranscriptionDelta",
    "StreamPartInputTranscriptionEnd",
    "StreamPartOutputTranscriptionStart",
    "StreamPartOutputTranscriptionDelta",
    "StreamPartOutputTranscriptionEnd",
    "StreamPartTurnStart",
    "StreamPartTurnEnd",
    "StreamPartStart",
    "StreamPartEnd",
    "StreamPart",
]



from ..utils.decorators import tool
from agentia.tools import Plugin, PluginInitError

# Plugins
from .skills import Skills
from .web import Web
from .system import System


__all__ = [
    "Plugin",
    "PluginInitError",
    "tool",
    "Skills",
    "Web",
    "System",
]



from dataclasses import dataclass


@dataclass
class LiveChunkAudio:
    data: bytes
    mime_type: str = "audio/pcm;rate=16000"


@dataclass
class LiveChunkVideo:
    data: bytes
    mime_type: str = "image/jpeg"


@dataclass
class LiveChunkImage:
    data: bytes
    mime_type: str = "image/jpeg"


@dataclass
class LiveChunkText:
    text: str


@dataclass
class LiveChunkEnd: ...


LiveChunk = (
    LiveChunkAudio | LiveChunkVideo | LiveChunkImage | LiveChunkText | LiveChunkEnd
)



"""Utility for downscaling images to standard resolutions."""

from typing import Literal
from io import BytesIO

ScaleOption = Literal["360p", "720p", "1080p", "raw"]

_TARGET_PIXELS: dict[ScaleOption, int] = {
    "360p": 640 * 360,  # 230_400
    "720p": 1280 * 720,  # 921_600
    "1080p": 1920 * 1080,  # 2_073_600
}


def downscale_jpeg(data: bytes, scale: ScaleOption) -> bytes:
    """Downscale a JPEG image to roughly match the target pixel count.

    The aspect ratio is preserved. The image is scaled so that
    ``new_w * new_h ≈ target_pixels``. If the image already has fewer
    pixels than the target, it is returned unchanged.

    Args:
        data: JPEG image bytes.
        scale: Target resolution or "raw" to skip scaling.

    Returns:
        JPEG image bytes, possibly resized.
    """
    if scale == "raw":
        return data

    import math
    from PIL import Image

    target_pixels = _TARGET_PIXELS[scale]
    img = Image.open(BytesIO(data))
    orig_w, orig_h = img.size
    orig_pixels = orig_w * orig_h

    if orig_pixels <= target_pixels:
        return data

    ratio = math.sqrt(target_pixels / orig_pixels)
    new_w = int(orig_w * ratio)
    new_h = int(orig_h * ratio)

    img = img.resize((new_w, new_h), Image.LANCZOS)  # type: ignore[attr-defined]
    buf = BytesIO()
    img.save(buf, format="JPEG")
    return buf.getvalue()



from typing import override
from ._openai_api import OpenAIAPIProvider
from .openrouter import OpenRouter as _OpenRouter
import os

# Model list: https://developers.openai.com/api/docs/models


class OpenAI(OpenAIAPIProvider):
    def __init__(
        self, model: str, api_key: str | None = None, base_url: str | None = None
    ):
        api_key = api_key or os.environ.get("OPENAI_API_KEY")
        if not api_key:
            raise ValueError("OPENAI_API_KEY environment variable not set")
        base_url = base_url or os.environ.get(
            "OPENAI_BASE_URL", "https://api.openai.com/v1"
        )
        super().__init__(name="openai", model=model, api_key=api_key, base_url=base_url)

    @override
    async def _fetch_context_length(self) -> int:
        # OpenAI's models endpoint doesn't include context length,
        # so we query OpenRouter's model catalog as a fallback.
        return await _OpenRouter.fetch_context_length(f"openai/{self.model}")



import inspect
from typing import Callable, overload


class EventEmitter[EventHandler: Callable]:
    def __init__(self):
        self._listeners: set[EventHandler] = set()

    @overload
    def on(self, listener: EventHandler): ...
    @overload
    def on(self) -> Callable[[EventHandler], EventHandler]: ...

    def on(self, listener: EventHandler | None = None):
        if listener is None:

            def decorator(func: EventHandler, /):
                self.on(func)
                return func

            return decorator
        self._listeners.add(listener)

    @overload
    def off(self, listener: EventHandler): ...
    @overload
    def off(self) -> Callable[[EventHandler], EventHandler]: ...

    def off(self, listener: EventHandler | None = None):
        if listener is None:

            def decorator(func: EventHandler, /):
                self.off(func)
                return func

            return decorator
        self._listeners.remove(listener)

    def __len__(self):
        return len(self._listeners)

    async def emit(self, *args, **kwargs):
        results = []
        for listener in self._listeners:
            r = listener(*args, **kwargs)
            if inspect.isawaitable(r):
                r = await r
            results.append(r)
        return results



import os
from typing import (
    TYPE_CHECKING,
    Callable,
    Literal,
    Optional,
    Sequence,
    overload,
    Any,
)
import logging
import uuid
from agentia.history import History
from agentia.live import LiveOptions
from agentia.llm import LLMOptions, get_provider
from agentia.llm.react import run_react_loop, run_react_loop_streamed
from agentia.llm.completion import ChatCompletion
from agentia.llm.stream import ChatCompletionStream, ChatCompletionEvents
from agentia.models.chat import ResponseFormatJson
from agentia.tools.tools import Tool, ToolSet
from agentia.models import (
    Message,
    ObjectType,
    Usage,
    UserConsentRequest,
    UserMessage,
)
from agentia.tools.mcp import MCPContext
from agentia.utils.commands import Commands
from agentia.utils.compact import compact_history
from agentia.utils.event_emitter import EventEmitter
from pathlib import Path
from agentia.live import Live

if TYPE_CHECKING:
    from agentia.plugins.skills import Skills


class AgentEvents:
    def __init__(self):
        self.end_of_turn = EventEmitter[Callable[[], Any]]()


class Agent:
    def __init__(
        self,
        model: str | None = None,
        tools: Sequence[Tool] | None = None,
        id: str | None = None,
        instructions: (
            str
            | Callable[[], str | None]
            | Sequence[str | Callable[[], str | None]]
            | None
        ) = None,
        options: LLMOptions | None = None,
        skills: Sequence[Path | str] | "Skills" | bool = False,
    ) -> None:
        from agentia.plugins.skills import Skills

        self.__id = str(id or uuid.uuid4())
        self.options = (
            LLMOptions(**options)
            if isinstance(options, dict)
            else (options or LLMOptions())
        )
        if skills is not False:
            tools = tools or []
            if any(isinstance(t, Skills) for t in tools):
                raise ValueError("Multiple Skills plugins provided.")
            if skills is True:
                tools = list(tools) + [Skills()]
            elif isinstance(skills, Skills):
                tools = list(tools) + [skills]
            elif isinstance(skills, Sequence):
                tools = list(tools) + [Skills(search_paths=skills)]
        self.tools = ToolSet(tools or [], self)
        if not model:
            model = os.getenv("AGENTIA_DEFAULT_MODEL", "openai/gpt-5-mini")
        self.__model = model
        self.provider = get_provider(model)
        self.history = History()
        if instructions:
            if isinstance(instructions, (str, Callable)):
                self.history.add_instructions(instructions)
            else:
                for i in instructions:
                    self.history.add_instructions(i)
        self.add_instructions(self.tools.get_instructions)
        self.log = logging.getLogger("agentia.agent")
        self._mcp_context: Optional[MCPContext] = None
        self._temp_mcp_context: Optional[MCPContext] = None
        self.events = AgentEvents()
        self.commands = Commands(self)
        self.__deferred_messages: list[UserMessage] = []
        self.__live: Optional[Live] = None

    @property
    def id(self) -> str:
        return self.__id

    @property
    def model(self) -> str:
        return self.__model

    @property
    def current_context_length(self) -> int:
        return self.history.current_tokens

    @property
    def usage(self) -> "Usage":
        return self.history.usage

    async def get_max_context_length(self) -> int:
        """Return the context length of the model, delegating to the provider."""
        return await self.provider.get_context_length()

    def add_instructions(self, instructions: str | Callable[[], str | None]) -> None:
        self.history.add_instructions(instructions)

    def __merge_options(self, options: LLMOptions | None) -> LLMOptions:
        options_merged = {}
        for k, v in self.options.model_dump().items():
            options_merged[k] = v
        if options:
            for k, v in options.model_dump().items():
                if v is not None:
                    options_merged[k] = v
        return LLMOptions(**options_merged)

    async def _auto_compact_if_needed(self, options: LLMOptions) -> None:
        DEFAULT_AUTO_COMPACT_THRESHOLD = 90
        if options.auto_compact:
            effort = options.auto_compact_effort or "low"
            length_threshold_percent = float(
                options.auto_compact_threshold or DEFAULT_AUTO_COMPACT_THRESHOLD
            )
            if (
                self.current_context_length
                / (await self.get_max_context_length())
                * 100
                > length_threshold_percent
            ):
                await self.compact(effort=effort)

    @overload
    def run(
        self,
        prompt: str | Message | Sequence[Message],
        *,
        stream: Literal[False] = False,
        events: Literal[False] = False,
        options: LLMOptions | None = None,
    ) -> ChatCompletion: ...

    @overload
    def run(
        self,
        prompt: str | Message | Sequence[Message],
        *,
        stream: Literal[True],
        events: Literal[False] = False,
        options: LLMOptions | None = None,
    ) -> ChatCompletionStream: ...

    @overload
    def run(
        self,
        prompt: str | Message | Sequence[Message],
        *,
        stream: Literal[True],
        events: Literal[True],
        options: LLMOptions | None = None,
    ) -> ChatCompletionEvents: ...

    def run(
        self,
        prompt: str | Message | Sequence[Message],
        *,
        stream: bool = False,
        events: bool = False,
        options: LLMOptions | None = None,
    ) -> ChatCompletion | ChatCompletionStream | ChatCompletionEvents:
        options_merged = self.__merge_options(options)
        if stream:
            x = run_react_loop_streamed(self, prompt, events, options_merged, None)
        else:
            assert not events, "events=True is only supported with stream=True"
            x = run_react_loop(self, prompt, options_merged, None)
        return x

    async def generate_object[T: ObjectType](
        self,
        prompt: str | Message | Sequence[Message],
        type: type[T],
        options: LLMOptions | None = None,
    ) -> T:
        options_merged = self.__merge_options(options)
        options_merged.response_format = ResponseFormatJson.from_model(type)
        result_msg = await run_react_loop(self, prompt, options_merged, None)
        return result_msg.parse(type)

    async def __aenter__(self):
        self._mcp_context = MCPContext()
        await self._mcp_context.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        assert self._mcp_context is not None
        await self._mcp_context.__aexit__(exc_type, exc_val, exc_tb)
        self._mcp_context = None

    async def user_consent(self, request: UserConsentRequest) -> bool | str | None:
        return True

    async def user_consent_guard(self, request: str | UserConsentRequest):
        if isinstance(request, str):
            request = UserConsentRequest(message=request)
        consent = await self.user_consent(request)
        if consent is True:
            return
        elif consent is False:
            raise PermissionError("User denied consent.")
        elif isinstance(consent, str):
            raise PermissionError(f"User denied consent: {consent}")
        elif consent is None:
            raise PermissionError("User did not respond to consent request.")
        else:
            raise ValueError(f"Invalid consent response: {consent}")

    async def defer(self, msg: UserMessage) -> None:
        """
        Enqueue a message to be added to the conversation history immediately after the current turn finishes.
        This can be used to add messages in the middle of a ReAct loop.
        """
        if self.__live is not None:
            await self.__live.send_message(msg)
        else:
            self.__deferred_messages.append(msg)

    def has_deferred_messages(self) -> bool:
        """
        Check if there are any messages enqueued with `defer()`.
        """
        return len(self.__deferred_messages) > 0

    def clear_deferred_messages(self) -> list[UserMessage]:
        """
        Clear any messages enqueued with `defer()` and return them.
        """
        msgs = self.__deferred_messages
        self.__deferred_messages = []
        return msgs

    async def compact(
        self,
        effort: Literal["low", "medium", "high"] = "low",
        model: str | None = None,
    ) -> None:
        await compact_history(self, effort, model)

    def live(self, options: LiveOptions | None = None) -> Live:
        if not self.provider.supports_live:
            raise NotImplementedError("This provider does not support live sessions")
        if self.__live is not None:
            raise RuntimeError("Live session already active")
        self.__live = Live(self, options)
        return self.__live



from . import Plugin, tool
from pathlib import Path
from typing import Any, Sequence, override
from pydantic import BaseModel
from agentia.utils.markdown import MarkdownDoc, load_markdown

INSTRUCTIONS_TEMPLATE = """\
# SKILLS

Skills provide specialized capabilities and domain knowledge.

When users ask you to perform tasks, check if any of the available skills match.

Use the `Skill` tool to execute a skill. Load only the necessary skills to complete the tasks.

## AVAILABLE SKILLS
{skills}
"""


class Skill(BaseModel):
    path: Path
    name: str
    description: str | None = None
    metadata: dict[str, Any]
    doc: MarkdownDoc | None = None

    @property
    def loaded(self) -> bool:
        return self.doc is not None

    @staticmethod
    def load(path: Path) -> "Skill":
        md = path / "SKILL.md"
        assert md.is_file(), f"Skill metadata file '{md}' does not exist"
        doc = load_markdown(str(md))
        if not ((name := doc.metadata.get("name")) and isinstance(name, str)):
            raise ValueError(f"Skill name must be a string in '{md}'")
        desc = doc.metadata.get("description", None)
        if desc is not None and not isinstance(desc, str):
            raise ValueError(f"Skill description must be a string in '{md}'")
        return Skill(
            path=path,
            name=name,
            description=desc,
            metadata=doc.metadata,
            doc=doc,
        )

    @property
    def disable_model_invocation(self) -> bool:
        return self.metadata.get("disable-model-invocation", False)

    @property
    def user_invocable(self) -> bool:
        return self.metadata.get("user-invocable", True)

    def execute(self, args: list[str] | None = None) -> str:
        if self.doc:
            return self.doc.content
        self.doc = load_markdown(
            str(self.path / "SKILL.md"), args=args, substitutions=["args", "bash"]
        )
        return self.doc.content


class Skills(Plugin):
    def __init__(self, search_paths: Sequence[Path | str] | None = None) -> None:
        if not search_paths:
            # default search path
            search_paths = [
                Path.cwd() / ".skills",
                Path.cwd() / ".agentia" / "skills",
                Path.home() / ".config" / "agentia" / "skills",
            ]
        self.search_paths = [Path(p) for p in search_paths]
        self.user_invocable_skills: dict[str, Skill] = {}
        self.agent_skills: dict[str, Skill] = {}

    @override
    async def init(self):
        all_skills = self.load_all_skills(self.search_paths)
        for skill in all_skills.values():
            if skill.user_invocable:
                self.user_invocable_skills[skill.name] = skill
            if not skill.disable_model_invocation:
                self.agent_skills[skill.name] = skill

    def load_all_skills(self, search_paths: list[Path]) -> dict[str, Skill]:
        skills: dict[str, Skill] = {}
        for path in search_paths:
            if not path.is_dir():
                continue
            for skill in self.discover_skills(path):
                skills[skill.name] = skill
        return skills

    def is_skill(self, path: Path) -> bool:
        # Skip hidden directories and files
        if path.name.startswith(".") or path.name.startswith("_"):
            return False
        # Skip directories without SKILL.md
        if not (path / "SKILL.md").is_file():
            return False
        return True

    def ignore_skill(self, skill: Skill) -> bool:
        return False

    def discover_skills(self, path: Path) -> list[Skill]:
        """Recursively discover skills from the given path."""
        assert path.is_dir(), f"'{path}' is not a directory"
        if self.is_skill(path):
            skill = Skill.load(path)
            if self.ignore_skill(skill):
                return []
            return [skill]
        skills = []
        for child in path.iterdir():
            if child.is_dir():
                skills.extend(self.discover_skills(child))
        return skills

    @override
    def get_instructions(self) -> str:
        s = ""
        for skill in self.agent_skills.values():
            if skill.disable_model_invocation:
                continue
            s += f"- **{skill.name}**:\n    * path: {str(skill.path)}\n    * description: {skill.description}\n\n"
        return INSTRUCTIONS_TEMPLATE.format(skills=s)

    @tool(name="Skill")
    def execute_skill(self, skill_name: str):
        """
        Execute a skill within the main conversation. All available skills are listed in system instructions.

        How to invoke:
            - Invoke the tool with `skill_name` set to the exact name of an available skill (no leading slash).

        Returns the content of the skill's SKILL.md after processing any directives with the provided arguments.
            - For file-references (`@path/to/file`, either absolute path or relative to the skill path), use the Read tool to access the file content.
            - Always use this tool to execute skills instead of loading SKILL.md directly.
        """
        skill = self.agent_skills.get(skill_name)
        if not skill:
            raise ValueError(f"Skill '{skill_name}' not found")
        if skill.disable_model_invocation:
            raise ValueError(f"Skill '{skill_name}' is disabled for model invocation")
        result_content = skill.execute([])
        return result_content



from typing import Any, AsyncGenerator, Generator
from agentia.llm.completion import async_gen_to_sync


def patch_asyncio():
    import nest_asyncio

    nest_asyncio.apply()


def patch_streamlit():
    import streamlit
    from streamlit import type_util

    if not hasattr(streamlit, "_async_generator_to_sync_patched"):

        def async_generator_to_sync(
            async_gen: AsyncGenerator[Any, Any],
        ) -> Generator[Any, Any, Any]:
            return async_gen_to_sync(async_gen)

        type_util.async_generator_to_sync = async_generator_to_sync
        setattr(streamlit, "_async_generator_to_sync_patched", True)


def patch_all():
    def try_patch(func):
        try:
            func()
        except ImportError:
            ...

    try_patch(patch_asyncio)
    try_patch(patch_streamlit)



from pathlib import Path
from typing import Annotated, Any, Literal
from pydantic import BaseModel, Field, JsonValue
from uuid import uuid4
import base64


type DataContent = bytes | str

type FinishReason = Literal[
    "stop",
    "length",
    "content-filter",
    "tool-calls",
    "interrupted",
    "error",
    "other",
    "unknown",
]


type ToolChoice = Literal["auto", "none", "required"] | str


class Usage(BaseModel):
    input_tokens: int | None = None
    output_tokens: int | None = None
    total_tokens: int | None = None
    reasoning_tokens: int | None = None
    cached_input_tokens: int | None = None

    def __add__(self, y: "Usage") -> "Usage":
        x = Usage(**self.model_dump())
        if y.input_tokens:
            x.input_tokens = (x.input_tokens or 0) + y.input_tokens
        if y.output_tokens:
            x.output_tokens = (x.output_tokens or 0) + y.output_tokens
        if y.total_tokens:
            x.total_tokens = (x.total_tokens or 0) + y.total_tokens
        if y.reasoning_tokens:
            x.reasoning_tokens = (x.reasoning_tokens or 0) + y.reasoning_tokens
        if y.cached_input_tokens:
            x.cached_input_tokens = (x.cached_input_tokens or 0) + y.cached_input_tokens
        return x


class FunctionTool(BaseModel):
    type: Literal["function"] = "function"
    name: str
    input_schema: JsonValue
    description: str | None = None

    def to_openai_schema(self) -> Any:
        return {
            "type": self.type,
            "name": self.name,
            "description": self.description,
            "parameters": self.input_schema,
            "strict": True,
        }


class ProviderDefinedTool(BaseModel):
    type: Literal["provider-defined"] = "provider-defined"
    id: str
    name: str
    args: dict[str, JsonValue] = Field(default_factory=dict)

    def to_openai_schema(self) -> Any:
        return {"type": self.id, **self.args}


type ToolSchema = Annotated[
    FunctionTool | ProviderDefinedTool, Field(discriminator="type")
]


class ToolCall(BaseModel):
    type: Literal["tool-call"] = "tool-call"

    tool_call_id: str
    """The identifier of the tool call. It must be unique across all tool calls."""

    tool_name: str
    """The name of the tool that should be called."""

    input: dict[str, JsonValue]
    """JSON object with the tool call arguments. Must match the parameters schema of the tool."""

    provider_executed: bool | None = None
    """
    Whether the tool call will be executed by the provider.
    If this flag is not set or is false, the tool call will be executed by the client.
    """


class File(BaseModel):
    media_type: str
    data: DataContent
    id: str = Field(default_factory=lambda: str(uuid4()))

    def to_url(self) -> str:
        if isinstance(self.data, bytes):
            return (
                f"data:{self.media_type};base64,{base64.b64encode(self.data).decode()}"
            )
        elif isinstance(self.data, str):
            if self.data.startswith(("http://", "https://", "data:")):
                return self.data
            else:
                path = Path(self.data)
                if not path.is_file():
                    raise ValueError(f"File path '{self.data}' does not exist")
                return f"data:{self.media_type};base64,{base64.b64encode(path.read_bytes()).decode()}"
        else:
            return str(self.data)

    def to_bytes(self) -> bytes:
        if isinstance(self.data, bytes):
            return self.data
        elif isinstance(self.data, str):
            if self.data.startswith("data:"):
                header, b64data = self.data.split(",", 1)
                return base64.b64decode(b64data)
            elif self.data.startswith(("http://", "https://")):
                raise ValueError("Cannot convert URL data to bytes")
            else:
                path = Path(self.data)
                if not path.is_file():
                    raise ValueError(f"File path '{self.data}' does not exist")
                return path.read_bytes()
        else:
            raise ValueError("Invalid data type for File")


class ToolCallResponse(BaseModel):
    type: Literal["tool-result"] = "tool-result"

    tool_call_id: str
    """The ID of the tool call that this result is associated with."""

    tool_name: str
    """Name of the tool that generated this result."""

    input: dict[str, JsonValue]

    output: JsonValue
    """Result of the tool call. This is a JSON-serializable object."""

    output_files: list[File] = Field(default_factory=list)
    """Optional list of files generated by the tool call."""

    is_error: bool | None = None
    """Optional flag if the result is an error or an error message."""

    provider_executed: bool | None = None
    """
    Whether the tool result was generated by the provider.
    If this flag is set to true, the tool result was generated by the provider.
    If this flag is not set or is false, the tool result was generated by the client.
    """

    preliminary: bool | None = None
    """
    Whether the tool result is preliminary.

    Preliminary tool results replace each other, e.g. image previews.
    There always has to be a final, non-preliminary tool result.

    If this flag is set to true, the tool result is preliminary.
    If this flag is not set or is false, the tool result is not preliminary.
    """


class Annotation(BaseModel):
    type: Literal["annotation"] = "annotation"

    end: int
    """The index of the last character of the URL citation in the message."""

    start: int
    """The index of the first character of the URL citation in the message."""

    title: str
    """The title of the web resource."""

    url: str
    """The URL of the web resource."""


class UserConsentRequest(BaseModel):
    type: Literal["user-consent-request"] = "user-consent-request"
    id: str = Field(default_factory=lambda: str(uuid4()))
    message: str
    details: str | None = None

    def __init__(self, message: str, details: str | None = None):
        super().__init__(message=message, details=details)


__all__ = [
    "DataContent",
    "FinishReason",
    "ToolChoice",
    "Usage",
    "FunctionTool",
    "ProviderDefinedTool",
    "ToolSchema",
    "ToolCall",
    "ToolCallResponse",
    "File",
    "Annotation",
    "UserConsentRequest",
]



import os
import re
from typing import TYPE_CHECKING, Any, Literal, Sequence
from openai import BaseModel
from pydantic import AnyUrl
from agentia import models

if TYPE_CHECKING:
    from agentia.llm.providers import Provider


class ReasoningOptions(BaseModel):
    enabled: bool | None = None
    effort: str | None = None
    exclude: bool | None = None
    max_tokens: int | None = None


class LLMOptions(BaseModel):
    max_output_tokens: int | None = None
    temperature: float | None = None
    stop_sequences: Sequence[str] | None = None
    top_p: float | None = None
    top_k: int | None = None
    presence_penalty: float | None = None
    frequency_penalty: float | None = None
    seed: int | None = None
    tool_choice: models.ToolChoice | None = None
    extra_body: dict[str, Any] | None = None
    response_format: models.ResponseFormat | type[BaseModel] | None = None
    parallel_tool_calls: bool | None = True
    reasoning: ReasoningOptions | None = None
    auto_compact: bool | None = None
    auto_compact_threshold: int | None = None
    auto_compact_effort: Literal["low", "medium", "high"] | None = None


def get_provider(selector: str) -> "Provider":
    # Parse selector
    if re.match(r"^[\w-]+:", selector) is None:
        # Default to openrouter if no provider specified
        default_provider = os.environ.get("AGENTIA_DEFAULT_PROVIDER", "openrouter")
        selector = f"{default_provider.strip().lower()}:{selector}"
    uri = AnyUrl(selector)
    for c in ["username", "password", "host", "port", "query", "fragment"]:
        if getattr(uri, c) is not None:
            raise ValueError(f"Invalid LLM selector: {selector}")
    provider, model = uri.scheme, (uri.path or "").strip("/")
    if not provider or not model:
        raise ValueError(f"Invalid LLM selector: {selector}")

    # Instantiate provider
    match provider:
        case "openai":
            from .providers.openai import OpenAI

            return OpenAI(model=model)
        case "vercel":
            from .providers.vercel import Vercel

            return Vercel(model=model)
        case "cloudflare":
            from .providers.cloudflare import Cloudflare

            return Cloudflare(model=model)
        case "openrouter":
            from .providers.openrouter import OpenRouter

            return OpenRouter(model=model)
        case "ollama":
            from .providers.ollama import Ollama

            return Ollama(model=model)
        case "gemini-live":
            from .providers.gemini_live import GeminiLive

            return GeminiLive(model=model)
        case _:
            raise ValueError(f"Unknown provider: {provider}")



from __future__ import annotations

import logging
from pathlib import Path

import dotenv

from .agent import Agent
from .models import *
from .utils.decorators import magic, ImageUrl
from .utils.decorators import tool
from .plugins import Plugin, PluginInitError
from .plugins.skills import Skills, Skill
from . import plugins
from .tools import ProviderTool, Tool, Tools, MCP, MCPContext, ToolResult
from .history import History
from .live import LiveOptions
from .llm import LLMOptions, ReasoningOptions
from .llm.completion import ChatCompletion
from .llm.stream import (
    TextStream,
    ReasoningStream,
    MessageStream,
    ChatCompletionStream,
    ChatCompletionEvents,
)

LOGGER = logging.getLogger("agentia")


def init_logging(level: logging._Level = logging.INFO):
    logging.basicConfig(
        level=level, format="[%(asctime)s][%(levelname)-8s][%(name)s] %(message)s"
    )
    logging.getLogger("httpx").setLevel(logging.WARNING)


def load_dotenv(
    *, dotenv_path: str | Path | None = None, cwd: bool = True, override: bool = True
):
    """Load environment variables from a .env file. By default, it looks for a .env file in the current working directory."""
    if dotenv_path:
        dotenv.load_dotenv(dotenv_path=dotenv_path, override=override)
    if cwd:
        dotenv.load_dotenv(dotenv_path=Path.cwd() / ".env", override=override)
    if not dotenv_path and not cwd:
        raise ValueError("Either dotenv_path must be provided or cwd must be True.")


__all__ = [
    # submodules
    "plugins",
    # agent
    "Agent",
    "History",
    # llm
    "ChatCompletion",
    "TextStream",
    "ReasoningStream",
    "MessageStream",
    "ChatCompletionStream",
    "ChatCompletionEvents",
    "LLMOptions",
    "ReasoningOptions",
    # models
    "DataContent",
    "FinishReason",
    "ToolChoice",
    "Usage",
    "ToolCall",
    "ToolCallResponse",
    "File",
    "Annotation",
    "UserConsentRequest",
    "MessageBase",
    "MessagePartBase",
    "MessagePartText",
    "MessagePartReasoning",
    "MessagePartFile",
    "MessagePartToolCall",
    "MessagePartToolResult",
    "MessagePart",
    "UserMessagePart",
    "AssistantMessagePart",
    "UserMessage",
    "ObjectType",
    "AssistantMessage",
    "ToolMessage",
    "Message",
    "ResponseFormatText",
    "ResponseFormatJson",
    "ResponseFormat",
    "StreamPartTextStart",
    "StreamPartTextDelta",
    "StreamPartTextEnd",
    "StreamPartReasoningStart",
    "StreamPartReasoningDelta",
    "StreamPartReasoningEnd",
    "StreamPart",
    # tools & plugins
    "Plugin",
    "PluginInitError",
    "Skills",
    "Skill",
    "magic",
    "tool",
    "ImageUrl",
    "ProviderTool",
    "Tool",
    "Tools",
    "ToolResult",
    "MCP",
    "MCPContext",
    # live
    "LiveOptions",
    # stream (live)
    "StreamPartAudioStart",
    "StreamPartAudioDelta",
    "StreamPartAudioEnd",
    "StreamPartInputTranscriptionStart",
    "StreamPartInputTranscriptionDelta",
    "StreamPartInputTranscriptionEnd",
    "StreamPartOutputTranscriptionStart",
    "StreamPartOutputTranscriptionDelta",
    "StreamPartOutputTranscriptionEnd",
    "StreamPartTurnStart",
    "StreamPartTurnEnd",
    # utils
    "init_logging",
    "load_dotenv",
]



from typing import override
from ._openai_api import OpenAIAPIProvider
from .openrouter import OpenRouter as _OpenRouter
import os

# Model list: https://developers.cloudflare.com/ai-gateway/usage/providers/


class Cloudflare(OpenAIAPIProvider):
    def __init__(self, model: str, api_key: str | None = None):
        api_key = (
            api_key
            or os.environ.get("CLOUDFLARE_AI_GATEWAY_API_KEY")
            or os.environ.get("CLOUDFLARE_API_KEY")
        )
        if not api_key:
            raise ValueError(
                "CLOUDFLARE_AI_GATEWAY_API_KEY or CLOUDFLARE_API_KEY environment variable not set"
            )
        account_id = os.environ.get("CLOUDFLARE_ACCOUNT_ID")
        if not account_id:
            raise ValueError("CLOUDFLARE_ACCOUNT_ID environment variable not set")
        base_url = f"https://gateway.ai.cloudflare.com/v1/{account_id}/default/compat"
        super().__init__(
            name="cloudflare", model=model, api_key=api_key, base_url=base_url
        )
        # self.extra_headers["cf-aig-authorization"] = f"Bearer {api_key}"

        self.extra_body["modalities"] = ["text", "image"]

    @override
    async def _fetch_context_length(self) -> int:
        # Cloudflare AI Gateway doesn't expose context length directly,
        # so we query OpenRouter's model catalog as a fallback.
        return await _OpenRouter.fetch_context_length(self.model)



import os
from . import Plugin, tool
from typing import Annotated, Literal
import httpx
from markdownify import markdownify
from tavily import AsyncTavilyClient


class Web(Plugin):
    def __init__(self, tavily_api_key: str | None = None):
        api_key = tavily_api_key or os.environ.get("TAVILY_API_KEY")
        if not api_key:
            raise ValueError("TAVILY_API_KEY is required for Web")
        self.__tavily = AsyncTavilyClient(api_key=api_key)

    async def __get(self, url: str):
        async with httpx.AsyncClient(
            headers={
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"
            }
        ) as client:
            res = await client.get(url)
            res.raise_for_status()
            content_type = res.headers.get("content-type")
            if content_type == "text/html":
                md = markdownify(res.text)
                return {"content": md, "content_type": content_type}
            # raw text: .txt, .csv, .json, .md
            if content_type in ["text/plain", "application/json", "text/markdown"]:
                return {"content": res.text, "content_type": content_type}
            # If the content is not supported, return the raw content
            return {"content": res.text, "content_type": content_type}

    @tool(name="WebSearch")
    async def web_search(
        self,
        query: Annotated[
            str, "The search query. Please be as specific and verbose as possible."
        ],
        topic: Annotated[
            Literal["general", "news", "finance"] | None,
            "The topic of the search. Default is 'general'.",
        ] = None,
        time_range: Annotated[
            Literal["day", "week", "month", "year"] | None,
            "The time range back from the current date based on publish date or last updated date.",
        ] = None,
        max_results: Annotated[
            int | None,
            "The maximum number of search results to return. Default is 5. It must be between 0 and 20.",
        ] = None,
    ):
        """
        Perform web search on the given query.
        Returning the top related search results in json format.
        When necessary, you need to combine this tool with the get_webpage_content tools (if available), to browse the web in depth by jumping through links.
        """

        tavily_results = await self.__tavily.search(
            query=query,
            search_depth="advanced",
            # max_results=10,
            include_answer=True,
            include_images=True,
            include_image_descriptions=True,
            topic=topic,  # type: ignore
            time_range=time_range,  # type: ignore
            max_results=max_results,  # type: ignore
        )
        return tavily_results

    @tool(name="WebFetch")
    async def get_webpage_content(
        self,
        url: Annotated[str, "The URL of the web page to get the content of"],
    ):
        """
        Access a web page by a URL, and fetch the content of this web page (in markdown format).
        You can always use this tool to directly access web content or access external sites.
        Use it at any time when you think you may need to access the internet.
        """
        result = await self.__tavily.extract(
            urls=url,
            # extract_depth="advanced",
            include_images=True,
        )
        failed_results = result.get("failed_results", [])
        if len(failed_results) > 0:
            try:
                return await self.__get(url)
            except Exception:
                pass
        return result



from contextlib import AsyncExitStack
from pathlib import Path
from typing import Any, Literal, Optional, overload
from typing import TYPE_CHECKING, Sequence
from fastmcp import Client
from fastmcp.mcp_config import (
    MCPConfig,
    CanonicalMCPServerTypes,
    StdioMCPServer,
    RemoteMCPServer,
)
from httpx import Auth
from mcp.types import Tool as FastMCPTool


if TYPE_CHECKING:
    from agentia.tools.tools import _MCPTool
    from agentia.agent import Agent


class MCPContext:
    def __init__(self):
        self.exit_stack = AsyncExitStack()
        self._servers: list[MCP] = []
        self.active = False

    async def __aenter__(self):
        self.active = True
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.active = False
        for server in self._servers:
            server.context_available = False
        await self.exit_stack.aclose()


class _ClientWrapper(Client):
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await super().__aexit__(exc_type, exc_val, exc_tb)
        await self.transport.close()


def _convert_tool_format(tool: FastMCPTool) -> Any:
    converted_tool = {
        "name": tool.name,
        "description": tool.description,
        "parameters": tool.inputSchema,
    }
    return converted_tool


class MCP:
    @overload
    def __init__(
        self,
        name: str,
        *,
        command: str,
        args: Sequence[str | Path],
        env: dict[str, str] | None = None,
        cwd: str | Path | None = None,
        timeout: int | None = None,
        type: Literal["local"] = "local",
        context: Optional[MCPContext] = None,
    ): ...

    @overload
    def __init__(
        self,
        name: str,
        *,
        type: Literal["http", "sse", "streamable-http"],
        url: str,
        headers: dict[str, str] | None = None,
        auth: str | Literal["oauth"] | Auth | None = None,
        timeout: int | None = None,
        context: Optional[MCPContext] = None,
    ): ...

    def __init__(
        self,
        name: str,
        *,
        command: str | None = None,
        args: Sequence[str | Path] | None = None,
        env: dict[str, str] | None = None,
        cwd: str | Path | None = None,
        url: str | None = None,
        headers: dict[str, str] | None = None,
        auth: str | Literal["oauth"] | Auth | None = None,
        timeout: int | None = None,
        type: Literal["sse", "http", "streamable-http", "local"] = "local",
        context: Optional[MCPContext] = None,
    ):

        self.cmd = None
        if command:
            self.cmd = [command] + ([a for a in args] if args else [])
        self.env = env
        self.cwd = cwd
        self.url = url
        self.headers = headers
        self.auth = auth
        self.type = type

        self.name = name
        self.initialized = False
        self.context_available = False
        self.__tools: list["_MCPTool"] = []
        self.session: Client | None = None
        self.timeout: int | None = timeout  # Maximum response time in milliseconds

        self.context = context

        def assert_no_local_args():
            assert command is None, "Remote MCP server does not support command"
            assert args is None, "Remote MCP server does not support args"
            assert env is None, "Remote MCP server does not support env"
            assert cwd is None, "Remote MCP server does not support cwd"

        match type:
            case "local":
                assert url is None, "Local MCP server does not support URL"
                assert headers is None, "Local MCP server does not support headers"
                assert auth is None, "Local MCP server does not support auth"
                assert command is not None and len(command) > 0, (
                    "Command must be specified"
                )
                assert args is not None, "Args must be specified"
            case "http" | "sse" | "streamable-http":
                assert_no_local_args()
                assert url is not None, "URL must be specified for HTTP server"

    def get_tools(self) -> list["_MCPTool"]:
        assert self.initialized, "MCP Server not initialized"
        return self.__tools

    async def init(self, agent: "Agent") -> None:
        from agentia.tools.tools import _MCPTool

        assert not self.initialized, "MCP Server already initialized"

        context: MCPContext
        if self.context:
            context = self.context
        elif agent._mcp_context and agent._mcp_context.active:
            context = agent._mcp_context
        elif agent._temp_mcp_context and agent._temp_mcp_context.active:
            context = agent._temp_mcp_context
        else:
            raise RuntimeError("MCP Server must be running in an MCP context")

        exit_stack = context.exit_stack
        self.context_available = True
        context._servers.append(self)

        config: CanonicalMCPServerTypes
        match self.type:
            case "local":
                assert self.cmd is not None
                config = StdioMCPServer(
                    command=str(self.cmd[0]),
                    args=[str(a) for a in self.cmd[1:]],
                    env=self.env or {},
                    cwd=str(self.cwd) if self.cwd else None,
                    timeout=self.timeout,
                )
            case "sse" | "streamable-http" | "http":
                assert self.url is not None
                config = RemoteMCPServer(
                    url=self.url,
                    transport=self.type,
                    headers=self.headers or {},
                    auth=self.auth,
                    timeout=self.timeout,
                )
            case _:
                raise ValueError(f"Unsupported MCP server type: {self.type}")

        self.session = await exit_stack.enter_async_context(
            _ClientWrapper(MCPConfig(mcpServers={self.name: config}))
        )

        tools = await self.session.list_tools()
        for tool in tools:
            schema = _convert_tool_format(tool)
            self.__tools.append(_MCPTool(name=tool.name, schema=schema, server=self))
        self.initialized = True

    async def run(self, tool: str, args: Any) -> Any:
        assert self.context_available, "Invalid or closed MCP context"
        assert self.session is not None, "MCP session not initialized"
        result = await self.session.call_tool(tool, args)
        assert self.context_available, "Invalid or closed MCP context"
        result_json = []
        for part in result.content:
            result_json.append(part.model_dump())
        return result_json



import abc
import re
from dataclasses import dataclass
from typing import TYPE_CHECKING, AsyncGenerator, Sequence
import httpx
from agentia.llm import LLMOptions
from agentia.models.base import ToolCallResponse
from agentia.models.chat import Message
from agentia.models.live import LiveChunk
from agentia.tools.tools import ToolSet
import agentia.models as models
from agentia.models.stream import StreamPart

if TYPE_CHECKING:
    from agentia.history import History
    from agentia.live import LiveOptions


@dataclass
class GenerationResult:
    message: models.AssistantMessage
    finish_reason: models.FinishReason
    usage: models.Usage


class Provider(abc.ABC):
    name: str
    supported_urls: dict[str, Sequence[re.Pattern | str]]
    """
    Supported URL patterns by media type for the provider.

    The keys are media type patterns or full media types (e.g. `*\\/*` for everything, `audio/*`, `video/*`, or `application/pdf`).
    and the values are arrays of regular expressions that match the URL paths.

    The matching should be against lower-case URLs.

    Matched URLs are supported natively by the model and are not downloaded.
    """

    def __init__(self, name: str, model: str):
        self.name = name
        self.__model = model
        self.supported_urls = {}
        self.__context_length: int | None = None

    @property
    def model(self) -> str:
        return self.__model

    async def get_context_length(self) -> int:
        """Return the context length of the model, fetching and caching it on first call."""
        if self.__context_length is None:
            self.__context_length = await self._fetch_context_length()
        return self.__context_length

    @abc.abstractmethod
    async def _fetch_context_length(self) -> int:
        """Fetch the context length from the provider API. Implemented by each provider."""
        ...

    @abc.abstractmethod
    async def generate(
        self,
        instructions: str,
        messages: list[Message],
        tools: ToolSet,
        options: LLMOptions,
        client: httpx.AsyncClient,
    ) -> GenerationResult: ...

    @abc.abstractmethod
    def stream(
        self,
        instructions: str,
        messages: list[Message],
        tools: ToolSet,
        options: LLMOptions,
        client: httpx.AsyncClient,
    ) -> AsyncGenerator[StreamPart, None]: ...

    # --- Live session methods (optional, for bidirectional providers) ---

    @property
    def supports_live(self) -> bool:
        return False

    async def connect_live(
        self,
        options: "LiveOptions",
        tools: ToolSet,
        instructions: str | None,
        history: "History",
    ) -> None:
        raise NotImplementedError("This provider does not support live sessions")

    async def disconnect_live(self) -> None:
        raise NotImplementedError("This provider does not support live sessions")

    async def send_live_chunk(self, chunk: LiveChunk) -> None:
        raise NotImplementedError("This provider does not support live sessions")

    async def send_tool_responses(self, responses: list[ToolCallResponse]) -> None:
        raise NotImplementedError("This provider does not support live sessions")

    async def receive(self) -> AsyncGenerator[StreamPart, None]:
        raise NotImplementedError("This provider does not support live sessions")
        # Make this a valid async generator
        yield  # type: ignore[misc]



from enum import Enum
import inspect
from typing import Annotated, Any, Literal, Sequence, Type
from .base import *
from openai.lib._parsing._completions import to_strict_json_schema  # type: ignore
from pydantic import BaseModel, Field, JsonValue
import json


class MessageBase(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)


class MessagePartBase(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)


class MessagePartText(MessagePartBase):
    type: Literal["text"] = "text"
    text: str

    def __init__(self, text: str, *, type: Literal["text"] = "text"):
        super().__init__(type=type, text=text)

    def to_openai_format(self) -> dict[str, Any]:
        return {"type": self.type, "text": self.text}


class MessagePartReasoning(MessagePartBase):
    type: Literal["reasoning"] = "reasoning"
    text: str

    def __init__(self, text: str, *, type: Literal["reasoning"] = "reasoning"):
        super().__init__(type=type, text=text)

    def to_openai_format(self) -> dict[str, Any]:
        return {"type": self.type, "text": self.text}


class MessagePartFile(MessagePartBase):
    type: Literal["file"] = "file"
    data: DataContent
    media_type: str
    filename: str | None = None

    def to_url(self) -> str:
        return File(data=self.data, media_type=self.media_type).to_url()

    def to_bytes(self) -> bytes:
        return File(data=self.data, media_type=self.media_type).to_bytes()

    def to_openai_format(self) -> dict[str, Any]:
        if self.media_type.startswith("image/"):
            url = self.to_url()
            return {"type": "image_url", "image_url": {"url": url}}
        if self.media_type.startswith("video/"):
            url = self.to_url()
            return {"type": "video_url", "video_url": {"url": url}}  # type: ignore
        elif self.media_type.startswith("audio/"):
            if isinstance(self.data, str) and (
                self.data.startswith(("http://", "https://"))
            ):
                raise ValueError("audio file parts with URLs")
            if isinstance(self.data, str):
                data = self.data
            else:
                assert isinstance(self.data, bytes)
                data = self.data.decode(encoding="utf-8")
            format: Literal["wav", "mp3"]
            if self.media_type == "audio/wav":
                format = "wav"
            elif self.media_type == "audio/mpeg" or self.media_type == "audio/mp3":
                format = "mp3"
            else:
                raise ValueError(f"audio file parts with media type {self.media_type}")
            return {
                "type": "input_audio",
                "input_audio": {"data": data, "format": format},
            }
        else:
            if isinstance(self.data, str) and self.data.startswith("file-"):
                # this is a file ID
                return {"type": "file", "file": {"file_id": self.data}}
            else:
                url = self.to_url()
                return {
                    "type": "file",
                    "file": {"filename": self.filename, "file_data": url},
                }


class MessagePartToolCall(MessagePartBase):
    type: Literal["tool-call"] = "tool-call"
    tool_call_id: str
    tool_name: str
    input: dict[str, JsonValue]
    provider_executed: bool | None = None


class MessagePartToolResult(MessagePartBase):
    type: Literal["tool-result"] = "tool-result"
    tool_call_id: str
    tool_name: str
    input: dict[str, JsonValue]
    output: JsonValue
    output_files: list[File] | None = None

    def serialize_output(self) -> str:
        return (
            json.dumps(self.output)
            if isinstance(self.output, (list, dict))
            else str(self.output)
        )


type MessagePart = Annotated[
    MessagePartText
    | MessagePartReasoning
    | MessagePartFile
    | MessagePartToolCall
    | MessagePartToolResult,
    Field(discriminator="type"),
]

type UserMessagePart = Annotated[
    MessagePartText | MessagePartFile, Field(discriminator="type")
]

type AssistantMessagePart = Annotated[
    MessagePartText | MessagePartReasoning | MessagePartToolCall,
    Field(discriminator="type"),
]


class UserMessage(MessageBase):
    role: Literal["user"] = "user"
    content: Sequence[UserMessagePart] | str

    def __init__(
        self,
        content: str | Sequence[UserMessagePart],
        *,
        role: Literal["user"] = "user",
    ):
        super().__init__(role=role, content=content)

    @property
    def content_list(self) -> list[UserMessagePart]:
        if isinstance(self.content, str):
            return [MessagePartText(text=self.content)]
        if not isinstance(self.content, (Sequence, list)):
            return [self.content]
        return list(self.content)

    @property
    def text(self) -> str:
        text = ""
        for part in self.content_list:
            if isinstance(part, MessagePartText):
                text += part.text
        return text

    def to_openai_format(self) -> dict[str, Any]:
        if isinstance(self.content, str):
            return {"role": self.role, "content": self.content}
        else:
            return {
                "role": self.role,
                "content": [part.to_openai_format() for part in self.content_list],
            }


type ObjectType = (
    BaseModel
    | Enum
    | str
    | int
    | float
    | bool
    | None
    | list["ObjectType"]
    | dict[str, "ObjectType"]
    | tuple["ObjectType", ...]
)


class _Result[X](BaseModel):
    result: X = Field(..., description="The result", title="Result")


class AssistantMessage(MessageBase):
    role: Literal["assistant"] = "assistant"
    content: Sequence[AssistantMessagePart] | str
    annotations: list[Annotation] | None = None

    def __init__(
        self,
        content: str | Sequence[AssistantMessagePart],
        *,
        role: Literal["assistant"] = "assistant",
        annotations: list[Annotation] | None = None,
    ):
        super().__init__(role=role, content=content)

    @property
    def content_list(self) -> list[AssistantMessagePart]:
        if isinstance(self.content, str):
            return [MessagePartText(text=self.content)]
        if not isinstance(self.content, Sequence):
            return [self.content]
        return list(self.content)

    @property
    def reasoning(self) -> str | None:
        text = ""
        for part in self.content_list:
            if isinstance(part, MessagePartReasoning):
                text += part.text
        return text

    @property
    def text(self) -> str:
        text = ""
        for part in self.content_list:
            if isinstance(part, MessagePartText):
                text += part.text
        return text

    @property
    def tool_calls(self) -> list[ToolCall]:
        tool_calls = []
        for part in self.content_list:
            if isinstance(part, MessagePartToolCall):
                tool_calls.append(
                    ToolCall(
                        tool_call_id=part.tool_call_id,
                        tool_name=part.tool_name,
                        input=part.input,
                        provider_executed=part.provider_executed,
                    )
                )
        return tool_calls

    def parse[T: ObjectType](self, return_type: type[T]) -> T:
        class Result(_Result[return_type]): ...

        if inspect.isclass(return_type) and issubclass(return_type, BaseModel):
            response_format = return_type
        else:
            response_format = Result

        json_string = ""
        for part in self.content:
            if isinstance(part, MessagePartText):
                json_string += part.text

        result = response_format.model_validate_json(json_string)
        if isinstance(result, _Result):
            return result.result
        else:
            return result

    def to_openai_format(self) -> dict[str, Any]:
        if isinstance(self.content, str):
            return {"role": self.role, "content": self.content}
        else:
            text_parts: list[str] = []
            images: list[Any] = []
            tool_calls: list[Any] = []
            reasoning = ""
            for i, p in enumerate(self.content_list):
                if p.type == "text":
                    text_parts.append(p.text)
                elif p.type == "reasoning":
                    reasoning += p.text
                elif p.type == "file":
                    if p.media_type.startswith("image/"):
                        images.append(
                            {"type": "image_url", "image_url": {"url": p.to_url()}}
                        )
                    else:
                        raise ValueError(
                            f"Only image files are supported in assistant messages, got media type {p.media_type}"
                        )
                elif p.type == "tool-call":
                    tool_calls.append(
                        {
                            "id": p.tool_call_id,
                            "type": "function",
                            "function": {
                                "name": p.tool_name,
                                "arguments": json.dumps(p.input),
                            },
                        }
                    )
            p = {
                "role": "assistant",
                "content": (
                    text_parts[0]
                    if len(text_parts) == 1
                    else text_parts
                    if text_parts
                    else None
                ),
                "tool_calls": tool_calls,
            }
            if reasoning:
                p["reasoning"] = reasoning
            if images:
                p["images"] = images
            return p


class ToolMessage(MessageBase):
    role: Literal["tool"] = "tool"
    content: Sequence[MessagePartToolResult]

    def __init__(
        self,
        content: Sequence[MessagePartToolResult],
        *,
        role: Literal["tool"] = "tool",
    ):
        super().__init__(role=role, content=content)

    def to_openai_format(self) -> dict[str, Any] | list[dict[str, Any]]:
        files: list[File] = []
        msgs = []
        for p in self.content:
            if p.output_files:
                p.output = {
                    "output": p.output,
                    "files_hint": f"The tool output includes {len(p.output_files)} file(s).",
                    "files": [f.id for f in p.output_files],
                }
                files.extend(p.output_files)
            val = p.serialize_output()
            msgs.append(
                {"role": "tool", "tool_call_id": p.tool_call_id, "content": val}
            )
        if files:
            user_msg = UserMessage(
                content=[
                    MessagePartText(text="[[TOOL_OUTPUT_FILES]]"),
                    *[
                        MessagePartFile(media_type=f.media_type, data=f.to_url())
                        for f in files
                    ],
                ]
            )
            msgs.append(user_msg.to_openai_format())
        return msgs if len(msgs) > 1 else msgs[0]


class CompactedMessage(MessageBase):
    role: Literal["compacted"] = "compacted"
    content: str
    original_messages: list["Message"]

    def to_openai_format(self) -> dict[str, Any]:
        return {
            "role": "assistant",
            "content": [
                {"type": "text", "text": "[[COMPACTED CONVERSATION SUMMARY]]"},
                {"type": "text", "text": self.content},
                {"type": "text", "text": "[[END OF CONVERSATION SUMMARY]]"},
            ],
        }

    def flattern(self) -> list[UserMessage | AssistantMessage | ToolMessage]:
        messages: list[UserMessage | AssistantMessage | ToolMessage] = []
        for msg in self.original_messages:
            if isinstance(msg, (UserMessage, AssistantMessage, ToolMessage)):
                messages.append(msg)
            elif isinstance(msg, CompactedMessage):
                messages.extend(msg.flattern())
        return messages


type Message = Annotated[
    UserMessage | AssistantMessage | ToolMessage | CompactedMessage,
    Field(discriminator="role"),
]


CompactedMessage.model_rebuild()


class ResponseFormatText(BaseModel):
    type: Literal["text"] = "text"


class ResponseFormatJson(BaseModel):
    type: Literal["json"] = "json"
    json_schema: JsonValue | None = None
    name: str | None = None
    description: str | None = None

    @staticmethod
    def from_model[T: ObjectType](return_type: Type[T]) -> "ResponseFormatJson":
        class Result(_Result[return_type]): ...

        if inspect.isclass(return_type) and issubclass(return_type, BaseModel):
            response_format = return_type
        else:
            response_format = Result
        return ResponseFormatJson(
            json_schema=to_strict_json_schema(response_format),
            name=return_type.__name__,
            description=f"JSON object matching the schema of {return_type.__name__}",
        )


type ResponseFormat = Annotated[
    ResponseFormatText | ResponseFormatJson, Field(discriminator="type")
]


__all__ = [
    "MessageBase",
    "MessagePartBase",
    "MessagePartText",
    "MessagePartReasoning",
    "MessagePartFile",
    "MessagePartToolCall",
    "MessagePartToolResult",
    "MessagePart",
    "UserMessagePart",
    "AssistantMessagePart",
    "UserMessage",
    "ObjectType",
    "AssistantMessage",
    "ToolMessage",
    "CompactedMessage",
    "Message",
    "ResponseFormatText",
    "ResponseFormatJson",
    "ResponseFormat",
]



import asyncio
import inspect
import textwrap
from typing import TYPE_CHECKING, Any, Callable, Optional, Sequence, Union

from pydantic import BaseModel, Field, JsonValue, ValidationError
from agentia.models.base import File
from agentia.tools.mcp import MCP
import agentia.models as models
from agentia.utils.decorators import ToolFuncParam
from inspect import Parameter
from pydantic.type_adapter import TypeAdapter

NAME_TAG = "agentia_tool_name"
DISPLAY_NAME_TAG = "agentia_tool_display_name"
IS_TOOL_TAG = "agentia_tool_is_tool"
DESCRIPTION_TAG = "agentia_tool_description"
METADATA_TAG = "agentia_tool_metadata"

if TYPE_CHECKING:
    from agentia.plugins import Plugin
    from agentia.agent import Agent


class _BaseTool:
    def __init__(
        self,
        name: str,
        schema: Any,
        display_name: str | None = None,
        description: str | None = None,
        metadata: Any | None = None,
    ):
        self.name = name
        self.display_name = display_name
        self.description = description
        self.schema = schema
        self.metadata = metadata


class _PythonFunctionTool(_BaseTool):
    def __init__(self, f: Callable[..., Any], plugin: Optional["Plugin"] = None):
        if plugin:
            name = getattr(f, NAME_TAG, None) or f"{plugin.name()}__{f.__name__}"
            display_name: str = (
                getattr(f, DISPLAY_NAME_TAG, None) or f"{plugin.name()}@{f.__name__}"
            )
        else:
            name = getattr(f, NAME_TAG, f.__name__)
            display_name = getattr(f, DISPLAY_NAME_TAG, name)
        description: str = textwrap.dedent(getattr(f, DESCRIPTION_TAG, f.__doc__) or "")
        metadata = getattr(f, METADATA_TAG, None)
        params = [
            ToolFuncParam(p, f.__name__)
            for p in inspect.signature(f).parameters.values()
            if p.name != "self"
        ]
        schema = _PythonFunctionTool.get_json_schema(
            name=name, description=description, fparams=params
        )
        super().__init__(
            name=name,
            schema=schema,
            display_name=display_name,
            description=description,
            metadata=metadata,
        )
        self.func = f
        self.plugin = plugin
        self.params = params

    @staticmethod
    def get_json_schema(
        name: str, description: str | None, fparams: list[ToolFuncParam]
    ) -> Any:
        params: Any = {"type": "object", "properties": {}, "required": []}
        for p in fparams:
            if p.required:
                params["required"].append(p.name)
            params["properties"][p.name] = p.schema
        return {
            "name": name,
            "description": description or "",
            "parameters": params,
        }

    def process_json_args(self, agent: "Agent", args: Any):
        from agentia.agent import Agent

        p_args: list[Any] = []
        kw_args: dict[str, Any] = {}

        def get_value(p: ToolFuncParam, name: str, default: Any, is_model: bool):
            if name not in args:
                return default
            if is_model:
                return p.type(**args[name])
            else:
                v = args[name]
                try:
                    return TypeAdapter(p.type).validate_python(v)
                except ValidationError as e:
                    raise ValueError(
                        f"Error parsing argument '{name}': {e}. Please retry with a valid value."
                    ) from e

        assert self.params is not None
        for p in self.params:
            is_model = inspect.isclass(p.type) and issubclass(p.type, BaseModel)
            match p.param.kind:
                case Parameter.POSITIONAL_ONLY if p.param.annotation == Agent:
                    if agent is not None:
                        p_args.append(agent)
                case Parameter.POSITIONAL_ONLY:
                    default = p.default if p.default != Parameter.empty else None
                    p_args.append(get_value(p, p.name, default, is_model))
                case Parameter.POSITIONAL_OR_KEYWORD | Parameter.KEYWORD_ONLY if (
                    p.param.annotation == Agent
                ):
                    kw_args[p.name] = agent
                case Parameter.POSITIONAL_OR_KEYWORD | Parameter.KEYWORD_ONLY:
                    default = p.default if p.default != Parameter.empty else None
                    kw_args[p.name] = get_value(p, p.name, default, is_model)
                case other:
                    raise ValueError(f"{other} is not supported")
        return p_args, kw_args


class _MCPTool(_BaseTool):
    def __init__(self, name: str, schema: Any, server: MCP):
        super().__init__(name=name, schema=schema)
        self.server = server


class ProviderTool(BaseModel):
    name: str
    args: dict[str, JsonValue] | None = None


type Tool = Plugin | Callable[..., Any] | ProviderTool | MCP
type Tools = Sequence[Tool]


class ToolResult(BaseModel):
    output: JsonValue
    files: list[File] = Field(default_factory=list)


class ToolSet:
    def __init__(self, tools: Tools, agent: "Agent"):
        from agentia.plugins import Plugin

        self.all_tools = tools
        self.plugins: dict[str, Plugin] = {}
        self.mcp_servers: dict[str, MCP] = {}
        self.agent = agent
        self.provider_tools: dict[str, ProviderTool] = {}
        self.__tools: dict[str, _BaseTool] = {}
        for t in tools:
            if isinstance(t, Plugin):
                self.__add_plugin(t)
            elif isinstance(t, MCP):
                self.__add_mcp_server(t)
            elif isinstance(t, ProviderTool):
                self.provider_tools[t.name] = t
            else:
                assert inspect.isfunction(t), (
                    "Expected a function, MCP server object, Plugin, or ProviderTool"
                )
                self.__add_function(t)
        self.__initialized = False

    async def init(self):
        if self.__initialized:
            return
        self.__initialized = True
        for plugin in self.plugins.values():
            try:
                plugin._agent = self.agent
                await plugin.init()
            except Exception as e:
                from . import PluginInitError

                raise PluginInitError(plugin.id(), e) from e
        for server in self.mcp_servers.values():
            await server.init(agent=self.agent)
            tools = server.get_tools()
            for tool in tools:
                self.__tools[tool.name] = tool

    def add(self, t: Union[Callable[..., Any], MCP, "Plugin"]) -> None:
        from agentia.plugins import Plugin

        if isinstance(t, Plugin):
            self.__add_plugin(t)
        elif isinstance(t, MCP):
            self.__add_mcp_server(t)
        else:
            assert inspect.isfunction(t), (
                "Expected a function, MCP server object, or Plugin"
            )
            self.__add_function(t)

    def __add_function(self, f: Callable[..., Any], plugin: Optional["Plugin"] = None):
        t = _PythonFunctionTool(f=f, plugin=plugin)
        self.__tools[t.name] = t

    def __add_plugin(self, p: "Plugin"):
        # Add all functions from the plugin
        for _n, method in inspect.getmembers(p, predicate=inspect.ismethod):
            if not getattr(method, IS_TOOL_TAG, False):
                continue
            self.__add_function(method, plugin=p)
        # Add the plugin to the list of plugins
        self.plugins[p.id()] = p

    def __add_mcp_server(self, server: MCP):
        self.mcp_servers[server.name] = server

    def get_plugin[T: "Plugin"](self, type: type[T]) -> Optional[T]:
        for p in self.plugins.values():
            if isinstance(p, type):
                return p
        return None

    def get_instructions(self) -> str | None:
        instructions = []
        for plugin in self.plugins.values():
            ins = plugin.get_instructions()
            if ins:
                if s := ins.strip():
                    instructions.append(s)
        if not instructions:
            return None
        return "\n\n".join(instructions)

    def is_empty(self) -> bool:
        return len(self.__tools) == 0

    def get_schema(self) -> Sequence[models.ToolSchema]:
        functions = [
            models.FunctionTool(
                name=v.name,
                description=v.description or "",
                input_schema=v.schema["parameters"],
            )
            for (k, v) in self.__tools.items()
        ]
        provider_tools = [
            models.ProviderDefinedTool(id=t.name, name=t.name, args=t.args or {})
            for t in self.provider_tools.values()
        ]
        return functions + provider_tools

    async def __run_python_tool(
        self,
        id: str,
        tool: _PythonFunctionTool,
        agent: "Agent",
        args: dict[str, JsonValue],
    ) -> models.ToolCallResponse:
        p_args, kw_args = tool.process_json_args(agent, args)
        output = tool.func(*p_args, **kw_args)  # type: ignore
        if inspect.isawaitable(output):
            output = await output
        files: list[File] = []
        if isinstance(output, ToolResult):
            files = output.files
            output = output.output
        return models.ToolCallResponse(
            tool_call_id=id,
            tool_name=tool.name,
            input=args,
            output=output,
            output_files=files,
        )

    async def __run_mcp_tool(
        self, id: str, tool: _MCPTool, args: dict[str, JsonValue]
    ) -> models.ToolCallResponse:
        result = await tool.server.run(tool.name, args)
        return models.ToolCallResponse(
            tool_call_id=id, tool_name=tool.name, input=args, output=result
        )

    async def __run_one(
        self, agent: "Agent", c: models.ToolCall
    ) -> models.ToolCallResponse:
        tool = self.__tools.get(c.tool_name, None)
        if not tool:
            raise ValueError(f"Tool {c.tool_name} not found")
        try:
            if isinstance(tool, _PythonFunctionTool):
                return await self.__run_python_tool(
                    id=c.tool_call_id,
                    tool=tool,
                    agent=agent,
                    args=c.input,
                )
            elif isinstance(tool, _MCPTool):
                return await self.__run_mcp_tool(
                    id=c.tool_call_id,
                    tool=tool,
                    args=c.input,
                )
            else:
                raise ValueError(f"Tool {c.tool_name} is of unknown type")
        except Exception as e:
            output: JsonValue = {"error": str(e)}
            r = models.ToolCallResponse(
                tool_call_id=c.tool_call_id,
                tool_name=c.tool_name,
                input=c.input,
                output=output,
            )
            return r

    async def run(
        self, agent: "Agent", tool_calls: list[models.ToolCall], parallel: bool
    ) -> list[models.ToolCallResponse]:
        if parallel:
            results = await asyncio.gather(
                *[self.__run_one(agent, c) for c in tool_calls]
            )
        else:
            results: list[models.ToolCallResponse] = []
            for c in tool_calls:
                results.append(await self.__run_one(agent, c))
        return results
